"""Retrieves data from Yahoo! Finance using yfinance module."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"

import datetime as dt
import typing as t
import zoneinfo

import yfinance as yf

import finquotes as fq

SOURCE = "finquotes.yfinance"


class YFinanceFeedMixin(fq.NetworkConfigMixin):
    """Retrieves data from Yahoo using yfinance."""

    def __init__(
        self: t.Self,
        as_of_date: dt.date | None = None,
        *,
        auto_adjust: bool = False,
        **kwargs: t.Any,
    ) -> None:
        """Initialize YFinanceFeedMixin."""
        super().__init__(**kwargs)
        self.as_of_date = as_of_date or fq.today()
        self.auto_adjust = auto_adjust

    @staticmethod
    def split_kwargs(kwargs: dict) -> tuple[dict, dict]:
        """Split kwargs between those known by YFinanceFeedMixin and everything else.

        Return:
        ------
            Tuple of yfinance keyword arguments, then the other keyword arguments.

        """
        yfinance_kwargs, kwargs = fq.NetworkConfigMixin.separate_network_kwargs(kwargs)
        yfinance_kwargs["as_of_date"] = kwargs.pop("as_of_date", None)
        yfinance_kwargs["auto_adjust"] = kwargs.pop("auto_adjust", False)
        return yfinance_kwargs, kwargs

    def mixin_fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        tzinfo: zoneinfo.ZoneInfo | None = None,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch current prices using yfinance."""
        frequency = frequency or "1d"
        end_date = end_date or self.as_of_date
        begin_date = begin_date or end_date
        assert isinstance(begin_date, dt.date), "begin_date must be a date"
        assert isinstance(end_date, dt.date), "end_date must be a date"
        tzinfo = tzinfo or fq.default_timezone()
        ticker = yf.Ticker(symbol)
        hist = ticker.history(
            auto_adjust=self.auto_adjust,
            start=begin_date,
            end=end_date + dt.timedelta(days=1),
            interval=frequency,
        ).rename_axis("date")
        # begin_date and end_date are used in dataframe query.
        begin_timestamp = dt.datetime.combine(  # noqa: F841
            begin_date,
            dt.time.min,
            tzinfo=tzinfo,
        )
        end_timestamp = dt.datetime.combine(  # noqa: F841
            end_date,
            dt.time.max,
            tzinfo=tzinfo,
        )
        filtered_hist = hist.query("@begin_timestamp <= date & date <= @end_timestamp")
        for index, row in filtered_hist.iterrows():
            yield fq.Price(
                symbol=symbol,
                date=index.date(),
                close=row["Close"],
                open_=row["Open"],
                high=row["High"],
                low=row["Low"],
                volume=row["Volume"],
                source=SOURCE,
            )

    def fetch_hist_splits(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Split, None, None]:
        """Fetch historical splits using yfinance."""
        ticker = yf.Ticker(symbol)
        splits = ticker.splits

        if splits.empty:
            fq.log.debug("no dividends found for %s", symbol)
            return

        # Filter splits by date range if necessary
        if begin_date or end_date:
            mask = True
            if begin_date:
                mask = mask & (splits.index.date >= begin_date)
            if end_date:
                mask = mask & (splits.index.date <= end_date)
            splits = splits[mask]

        for date, ratio in splits.items():
            # Convert from pandas Timestamp to date object
            ex_date = date.date() if hasattr(date, "date") else date
            # Create a Split object using the from_ratio method
            yield fq.Split.from_ratio(symbol, ex_date, fq.to_decimal(str(ratio)))

    def fetch_hist_dists(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Distribution, None, None]:
        """Fetch historical distributions (dividends) using yfinance."""
        ticker = yf.Ticker(symbol)
        dividends = ticker.dividends

        if dividends.empty:
            fq.log.debug("no dividends found for %s", symbol)
            return

        # Filter dividends by date range if necessary
        if begin_date or end_date:
            mask = True
            if begin_date:
                mask = mask & (dividends.index.date >= begin_date)
            if end_date:
                mask = mask & (dividends.index.date <= end_date)
            dividends = dividends[mask]

        for date, amount in dividends.items():
            # Convert from pandas Timestamp to date object
            ex_date = date.date() if hasattr(date, "date") else date
            yield fq.Distribution(
                symbol=symbol,
                ex_date=ex_date,
                amount=fq.to_decimal(str(amount)),
            )

    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch current prices using yfinance."""
        date = date or self.as_of_date
        tickers = yf.Tickers(symbols)
        prices = tickers.history(
            auto_adjust=self.auto_adjust,
            group_by="ticker",
            start=date - dt.timedelta(days=1),
            progress=False,
        )
        try:
            today_prices = prices.loc[f"{date:%Y-%m-%d}"]
        except KeyError as ex:
            msg = f"No data for {date} and {', '.join(symbols)}"
            raise ValueError(msg) from ex
        for symbol in symbols:
            row = today_prices.loc[symbol]
            yield fq.Price(
                symbol=symbol,
                date=date,
                close=row["Close"],
                open_=row["Open"],
                high=row["High"],
                low=row["Low"],
                volume=row["Volume"],
                source=SOURCE,
            )

    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[fq.Security, None, None]:
        """Fetch security info using yfinance."""
        for symbol in symbols:
            ticker = yf.Ticker(symbol)
            info = ticker.info
            yield fq.Security(
                symbol=symbol,
                name=info["shortName"],
                type=info["quoteType"].lower(),
                currency=info["currency"],
                exchange=fq.Exchange("X" + info["exchange"]),
                sector=info["sector"],
                industry=info["industry"],
                country=info["country"],
                source=SOURCE,
            )


class YFinanceHistFeed(
    YFinanceFeedMixin,
    fq.AbstractHistoricalFeed,
    fq.AbstractSplitFeed,
    fq.AbstractDistributionFeed,
):
    """Implements finquotes.CloseFeed with YFinanceFeedMixin."""

    def __init__(self: t.Self, /, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        splits: t.Sequence[fq.Split] | None = None,  # noqa: ARG002
        distributions: t.Sequence[fq.Distribution] | None = None,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch historical prices using yfinance."""
        return YFinanceFeedMixin.mixin_fetch_hist_prices(
            self,
            symbol,
            end_date,
            begin_date,
            frequency,
        )

    def fetch_hist_splits(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Split, None, None]:
        """Fetch historical splits using yfinance."""
        return YFinanceFeedMixin.fetch_hist_splits(self, symbol, end_date, begin_date)

    def fetch_hist_dists(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Distribution, None, None]:
        """Fetch historical distributions using yfinance."""
        return YFinanceFeedMixin.fetch_hist_dists(self, symbol, end_date, begin_date)


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> YFinanceHistFeed:
    """Forward to build_feed."""
    return YFinanceHistFeed(*args, **kwargs)


def build_split_feed(*args: t.Any, **kwargs: t.Any) -> YFinanceHistFeed:
    """Forward to build_feed."""
    return YFinanceHistFeed(*args, **kwargs)


def build_distribution_feed(*args: t.Any, **kwargs: t.Any) -> YFinanceHistFeed:
    """Forward to build_feed."""
    return YFinanceHistFeed(*args, **kwargs)


class YFinanceCloseFeed(YFinanceFeedMixin, fq.AbstractCloseFeed):
    """Implements finquotes.CloseFeed with YFinanceFeedMixin."""

    def __init__(self: t.Self, /, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
        **kwargs: t.Any,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch close prices using yfinance."""
        yield from YFinanceFeedMixin.fetch_close_prices(self, symbols, date, **kwargs)


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> YFinanceCloseFeed:
    """Forward to build_feed."""
    return YFinanceCloseFeed(*args, **kwargs)


class YFinanceSecurityFeed(YFinanceFeedMixin, fq.AbstractSecurityFeed):
    """Implements finquotes.SecurityFeed with YFinanceFeedMixin."""

    def __init__(self: t.Self, /, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[fq.Security, None, None]:
        """Fetch security information for the given symbols.

        :param symbols: An iterable of symbol strings to fetch security data for.

        Yields Security objects containing information about each symbol.

        """
        return YFinanceFeedMixin.fetch_securities(self, symbols)


def build_security_feed(*args: t.Any, **kwargs: t.Any) -> YFinanceSecurityFeed:
    """Forward to build_feed."""
    return YFinanceSecurityFeed(*args, **kwargs)



"""Retrieves data from Yahoo! Finance."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"


import datetime as dt
import functools
import typing as t
from collections.abc import Generator, Sequence
from decimal import Decimal as D  # noqa: N817

import requests

import finquotes as fq

get_api_key = functools.partial(fq.get_api_key, api_name="quandl")
SOURCE = "finquotes.quandl"


class QuandlFeedMixin(fq.NetworkConfigMixin):
    """Finquotes feed to download data from Quandl.

    See http://www.quandl.com

    You will need and API key to be able to use that feed.

    Implements both historical and close price retrieval interface.

    """

    base_url = "https://www.quandl.com/api/v3"
    date_fmt = "%Y-%m-%d"

    def __init__(
        self: t.Self,
        data_set: str | None = None,
        api_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize quandl API.

        :param api_key: Quandl API key.
        :type api_key: str or None
        :param str dataset: Date set of interest (defaults to WIKI)

        If not specified the API key will be looked up in the
        FINQUOTES_QUANDL_API_KEY variable or in the file apikeys.txt.

        See also: `finquotes.quandl.get_api_key`

        """
        super().__init__(**kwargs)
        self.api_key = api_key
        if self.api_key is None:
            self.api_key = get_api_key()
        self.data_set = data_set or "WIKI"

    def fetch_hist_price(self: t.Self, symbol: str, line: str) -> fq.Price:
        """Convert line returned by quandl to HistoricalPrice."""
        date_str, open_, high, low, close, volume, *_ = line.split(",")
        date = dt.date.fromisoformat(date_str)
        return fq.Price(
            symbol,
            date,
            D(close),
            open_=D(open_),
            low=D(low),
            high=D(high),
            volume=fq.to_int_or_none(volume),
            source=SOURCE,
        )

    def fetch_close_price(self: t.Self, line: str) -> fq.Price:
        """Convert line returned by quandl to HistoricalPrice."""
        symbol, date_str, open_, high, low, close, volume, *_ = line.split(",")
        date = dt.date.fromisoformat(date_str)
        return fq.Price(
            symbol,
            date,
            D(close),
            open_=D(open_),
            low=D(low),
            high=D(high),
            volume=fq.to_int_or_none(volume),
            source=SOURCE,
        )


class QuandlHistFeed(QuandlFeedMixin, fq.AbstractHistoricalFeed):
    """Historical feed from Quandl."""

    def __init__(
        self: t.Self,
        data_set: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize quandl API.

        :param api_key: Quandl API key.
        :param str dataset: Date set of interest (defaults to WIKI)

        If not specified the API key will be looked up in the
        FINQUOTES_QUANDL_API_KEY variable or in the file apikeys.txt.

        See also: `finquotes.quandl.get_api_key`

        """
        data_set = data_set or "WIKI"
        super().__init__(data_set=data_set, name=__name__, source=SOURCE + "/" + data_set, **kwargs)

    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,  # noqa: ARG002
        splits: t.Sequence[fq.Split] | None = None,  # noqa: ARG002
        distributions: t.Sequence[fq.Distribution] | None = None,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve prices from  Quandl.

        :param str symbol: symbol to get prices for
        :param datetime.date begin_date: start date of the serie
        :param datetime.date end_date: end date of the serie

        If end_date is not specified, it defaults to today. If begin_date is
        not specified, it defaults to end_date.

        """
        assert isinstance(symbol, str), f"expected str, not {type(symbol)}"
        if not end_date:
            end_date = fq.today()
        if not begin_date:
            begin_date = end_date
        params = (
            f"start_date={begin_date.strftime(self.date_fmt)}"
            f"&end_date={end_date.strftime(self.date_fmt)}"
            f"&api_key={self.api_key}"
        )
        url = f"{self.base_url}/datasets/{self.data_set}/{symbol}/data.csv?{params}"
        fq.log.debug("fetching %s...", url)
        response = requests.get(url, timeout=self.timeout)
        for line in response.text.split("\n"):
            if not line or line.startswith("Date"):
                continue
            if line.startswith("code"):
                raise ValueError(response.text.split("\n")[1])
            try:
                price = self.fetch_hist_price(symbol, line)
            except ValueError as err:
                fq.log.warn("failed to process %s for %s: %s", line, symbol, err)
                raise
            yield price


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> QuandlHistFeed:
    """Build finquotes.quandl.HistQuandlFeedMixin."""
    return QuandlHistFeed(*args, **kwargs)


class QuandlCloseFeed(QuandlFeedMixin, fq.AbstractCloseFeed):
    """Close feed from Quandl."""

    def __init__(
        self: t.Self,
        as_of_date: dt.date | None = None,
        api_key: str | None = None,
        data_set: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize quandl API.

        :param as_of_date: as of date for retrieval.
        :param api_key: Quandl API key.
        :param str dataset: Date set of interest (defaults to WIKI)

        If not specified the API key will be looked up in the
        FINQUOTES_QUANDL_API_KEY variable or in the file apikeys.txt.

        See also: `finquotes.quandl.get_api_key`

        """
        data_set = data_set or "WIKI"
        super().__init__(
            api_key=api_key,
            data_set=data_set,
            name=__name__,
            source=SOURCE + "/" + data_set,
            as_of_date=as_of_date,
            **kwargs,
        )

    def fetch_close_prices(
        self: t.Self,
        symbols: Sequence[str],
        date: dt.date | None = None,  # noqa: ARG002
        **kwargs: t.Any,  # noqa: ARG002
    ) -> Generator[fq.Price, None, None]:
        """Retrieve close prices for context as of date.

        :param symbols: symbols whose price is to be retrieved.
        :param date: date for which we want the prices.

        """
        not_found = set(symbols)
        symbols_str = ",".join(symbols)
        params = {
            "date": self.as_of_date.strftime(self.date_fmt),
            "ticker": symbols_str,
            "api_key": self.api_key,
        }
        url = f"{self.base_url}/datatables/{self.data_set}/PRICES.csv"
        response = requests.get(url, params=params, timeout=self.timeout)
        for line in response.text.split("\n"):
            if line.startswith("ticker"):
                continue
            if line.startswith("code"):
                quandl_msg = response.text.split("\n")[1]
                raise RuntimeError(quandl_msg)
            if not line:
                continue
            price = self.fetch_close_price(line)
            not_found.remove(price.symbol)
            yield price
        fq.log.warning("no price found for: %s", ",".join(not_found))


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> QuandlCloseFeed:
    """Build finquotes.quandl.QuandlFeedMixin."""
    return QuandlCloseFeed(*args, **kwargs)



"""Factor http connectivity.

Internal module that wraps requests and requests_html.

Named localhttp to avoid name conflict with requests.

"""

import logging
import typing as t

import requests
import requests.adapters
import requests_cache
from urllib3.util import retry

log = logging.getLogger("finquotes")


cached_session = requests_cache.CachedSession(
    "finquotes",
    backend="filesystem",
    expire_after=3600,
    use_cache_dir=True,
)


def set_retries(session: requests.Session, **kwargs: t.Any) -> None:
    """Update session with retry policy.

    **kwargs are passed to the urllib3.Retry constructor:
    https://urllib3.readthedocs.io/en/stable/reference/urllib3.util.html#urllib3.util.Retry

    """
    if "total" not in kwargs:
        kwargs["total"] = 5
    if "backoff_factor" not in kwargs:
        kwargs["backoff_factor"] = 1
    if "status_forcelist" not in kwargs:
        kwargs["status_forcelist"] = [502, 503, 504]
    retry_policy = retry.Retry(**kwargs)
    http_adapter = requests.adapters.HTTPAdapter(max_retries=retry_policy)
    session.mount("http://", http_adapter)
    session.mount("https://", http_adapter)


def request(
    session: requests.Session,
    *args: t.Any,
    **kwargs: t.Any,
) -> requests.Response:
    """Forward call to requests and return the response itself.

    The only additional argument is a session.

    :param session: requests session.

    From requests.Request:
    :param method: HTTP method to use.
    :param url: URL to send.
    :param headers: dictionary of headers to send.
    :param files: dictionary of {filename: fileobject} files to multipart upload.
    :param data: the body to attach to the request. If a dictionary or list of tuples [(key, value)] is provided,
        form-encoding will take place.
    :param params: URL parameters to append to the URL. If a dictionary or list of tuples [(key, value)] is provided,
        form-encoding will take place.
    :param auth: Auth handler or (user, pass) tuple.
    :param cookies: dictionary or CookieJar of cookies to attach to this request.
    :param hooks: dictionary of callback hooks, for internal usage.
    :param json: json for the body to attach to the request (if files or data is not specified).

    """
    log.debug("args=%s", args)
    log.debug("kwargs=%s", kwargs)
    req = requests.Request(*args, **kwargs)
    prepared = req.prepare()
    log.info("%s %s ...", prepared.method, prepared.url)
    assert session is not None, "session is None!"
    response = session.send(prepared)
    log.debug("response.status_code=%s", response.status_code)
    response.raise_for_status()
    return response


def get(
    session: requests.sessions.Session,
    url: str,
    params: dict[str, str] | None = None,
) -> str:
    """Log url being fetched with requests.get and return response.text.

    The session is generated on the fly.

    :param session: requests session.
    :param url: url being looked up.
    :param params: parameters passed wit the query.
    :param session: requests session, one will be created if none passed.

    :return: Response from the request.

    """
    req = requests.Request("GET", url, params=params)
    prepared = req.prepare()
    log.info("fetching %s ...", prepared.url)
    response = session.send(prepared)
    response.raise_for_status()
    return response.text



"""Command line interface useful for testing, forwards to cmdline."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"

import sys

from finquotes import cmdline

if __name__ == "__main__":
    cmdline.main(sys.argv[1:])  # pylint: disable=no-value-for-parameter



"""CSV module.

This module provides a simple interface to ingest prices from CSV files.

Was renamed csvdata from csv because of unresolved conflicts with the builtin csv module.
"""

import csv
import datetime as dt
import sys
import typing as t

from dateutil import parser

import finquotes as fq

SOURCE = "finquotes.csv"


class CsvCloseFeed(fq.NetworkConfigMixin, fq.AbstractCloseFeed):
    """Read close prices from a CSV input.

    Expected input is expected to contain the following columns not necessarily in this order
    but the name of the columns is expected to be as shown:
    - symbol: security ticker or symbol
    - date: date in YYYY-MM-DD format
    - open: open price
    - high: high price
    - low: low price
    - close: close price
    - volume: volume

    """

    def __init__(
        self: t.Self,
        input_: t.IO[str] | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the CSV feed."""
        super().__init__(name=__name__, source=SOURCE, **kwargs)
        if not input_:
            input_ = sys.stdin
        self.input = input_

    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
        **kwargs: t.Any,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch the close prices for the given symbols.

        See fq.AbstractCloseFeed.fetch_close_prices for more details.

        """
        symbols_of_interest = set(symbols)
        lines = [line.rstrip("\r\n") for line in self.input or []]
        csv_reader = csv.DictReader(lines)
        for row in csv_reader:
            if row["symbol"] in symbols_of_interest:
                date = parser.parse(row["date"]).date()
                yield fq.Price(
                    symbol=row["symbol"],
                    date=date,
                    close=row["close"],
                    open_=row["open"],
                    high=row["high"],
                    low=row["low"],
                    volume=fq.to_int_or_none(row["volume"]),
                    source=self.source,
                )

    def to_price(self: t.Self, line: str) -> fq.Price:
        """Convert price as string to a price."""
        symbol, date_str, open_, high, low, close, volume = line.split(",")
        date = dt.date.fromisoformat(date_str)
        return fq.Price(
            symbol,
            date,
            close,
            open_=open_,
            low=low,
            high=high,
            volume=int(volume),
            source=self.source,
        )


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> CsvCloseFeed:
    """Create an EodDataCloseFeed object."""
    return CsvCloseFeed(*args, **kwargs)



"""Retrieves data from http://www.morningstar.com.

Currently only security names. Need to leverage undocumented API for historical
prices and distributions.

"""

__author__ = "Jerome Lecomte"
__license__ = "MIT"


import typing as t

import bs4  # type: ignore[import-untyped]
import requests

import finquotes as fq

SOURCE = "finquotes.morningstar"


def html_to_security(html: str, source: str) -> fq.Security:
    """Extract security information from html.

    :param html: html page retrieved from Morningstar.
    :returns: finquotes.Security.

    """
    soup = bs4.BeautifulSoup(html, "html.parser")
    element = soup.find("h1")
    if element is None:
        msg = "element not found in html"
        raise AttributeError(msg)
    parent_element = element.parent
    if parent_element is None:
        msg = "parent element not found in html"
        raise AttributeError(msg)
    name, symbol = parent_element.text.split("\xa0")
    return fq.Security(source=source, symbol=symbol.strip(), name=name.strip())


class MorningstarSecurityFeed(fq.NetworkConfigMixin, fq.AbstractSecurityFeed):
    """Retrieves security data from morningstar.com."""

    url = "http://financials.morningstar.com/company-profile/c.action?t="

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[fq.Security, None, None]:
        """Fetch html data from the url and extract security information.

        :param symbols: security symbols.
        :type symbols: list of str.

        :rtype: yield finquotes.Security.

        """
        assert iter(symbols)
        assert not isinstance(symbols, str)
        with requests.Session() as session:
            for symbol in symbols:
                security = self.fetch_security(session, symbol)
                if security:
                    yield security

    def fetch_security(
        self: t.Self,
        session: requests.Session,
        symbol: str,
    ) -> fq.Security | None:
        """Retrieve and return a finquotes.Security from a url.

        :param session: requests session.
        :type session: requests.Session.
        :param symbol: symbol to fetch.
        :type symbol: str.
        :returns: finquotes.Security for that url.

        """
        url = self.url + symbol
        try:
            request = session.get(url, allow_redirects=True)
            request.raise_for_status()
            return html_to_security(request.text, self.source)
        except requests.exceptions.RequestException as error:
            fq.log.warning("fail to retrieve %s: %s", url, str(error))
        except AttributeError as error:
            fq.log.warning("%s not found in %s: %s", symbol, url, str(error))
        return None


def build_security_feed() -> MorningstarSecurityFeed:
    """Create and return a MorningstarSecurityFeed object."""
    return MorningstarSecurityFeed()



"""Retrieves data from various financial sources."""

import abc
import csv
import dataclasses
import datetime as dt
import decimal
import enum
import functools
import importlib
import importlib.metadata
import importlib.resources
import importlib.resources.abc
import io
import logging
import os
import typing as t
import zoneinfo
from decimal import Decimal as D  # noqa: N817
from pathlib import Path

import urllib3.exceptions

from . import localhttp
from .network_config import NetworkConfigMixin as NetworkConfigMixin

__version__ = importlib.metadata.version("finquotes")
log = logging.getLogger("finquotes")
# See Python cookbook, 3rd Edition, 13.12.
log.addHandler(logging.NullHandler())


# Configure where the API keys are found
api_key_file = Path("~/.finquotes/api_keys.txt").expanduser()


def default_timezone() -> zoneinfo.ZoneInfo:
    """Return default to EST time zone because prices are usually available in US markets.

    Will eventually be replaced wit a more flexible solution.

    """
    return zoneinfo.ZoneInfo("America/New_York")


def default_timezone_offset() -> dt.timedelta | None:
    """Return default timezone offset for EST time zone."""
    return default_timezone().utcoffset(dt.datetime.now(default_timezone()))


def default_timezone_offset_str() -> str:
    """Return the default timezone offset as a string."""
    offset = default_timezone_offset()
    if offset is None:
        return "+00:00"  # Default to UTC if no offset is available
    total_seconds = int(offset.total_seconds())
    hours = total_seconds // 3600  # Convert seconds to hours
    minutes = (total_seconds % 3600) // 60
    sign = "-" if total_seconds < 0 else "+"
    return f"{sign}{abs(hours):02d}:{abs(minutes):02d}"


def get_api_key(api_name: str, encoding: str | None = None) -> str:
    """Retrieve API key.

    :param api_name: name of the API
    :param encoding: encoding to use when reading the API key file, defaults to utf-8

    :return: api key for that api name.

    :raises KeyError: if the API key is not found
    """
    encoding = encoding or "utf-8"
    uc_api_name = api_name.replace("finquotes.", "").upper()
    api_key_env_var = f"FINQUOTES_{uc_api_name}_API_KEY"
    api_key_from_env = os.getenv(api_key_env_var)
    if api_key_from_env:
        log.info("using api key in environment variable %s", api_key_env_var)
        return api_key_from_env
    log.debug("cannot find api key in environment %s", api_key_env_var)
    try:
        with Path(api_key_file).open(encoding=encoding) as fh:
            lines = fh.read()
            for line in lines.split("\n"):
                if "=" not in line:
                    continue
                api, key = line.split("=")
                if api == api_name:
                    log.info("using api key in file %s", api_key_file)
                    return key
    except (ValueError, TypeError) as error:
        log.debug("failed to read api key from %s: %s", api_key_file, error)
    except FileNotFoundError as error:
        log.debug("failed to open %s for reading: %s", api_key_file, error)
    log.error(
        "failed to find an api key for api %s. Try to set %s or update the file %s with a value for %s=<your key>",
        api_name,
        api_key_env_var,
        api_key_file,
        api_name,
    )
    raise KeyError("api key for " + api_name + " not found or empty!")


def now() -> dt.datetime:
    """Return current timestamp as tz-aware datetime.datetime."""
    return dt.datetime.now(default_timezone())


def today() -> dt.date:
    """Return today's date."""
    return now().date()


def compute_default_dates(
    end_date: dt.date | None,
    begin_date: dt.date | None,
) -> tuple[dt.date, dt.date]:
    """Defaults the dates if they are different from None."""
    if end_date is None:
        end_date = today()
    if begin_date is None:
        begin_date = end_date
    assert isinstance(
        begin_date,
        dt.date,
    ), f"begin_date: date expected, got {type(begin_date)}"
    if isinstance(begin_date, dt.datetime):
        begin_date = begin_date.date()
    assert isinstance(
        end_date,
        dt.date,
    ), f"end_date: date expected, got {type(end_date)}"
    if isinstance(end_date, dt.datetime):
        end_date = end_date.date()
    assert begin_date <= end_date, (
        f"begin_date {begin_date:%Y-%m-%d} must be earlier than end_date {end_date:%Y-%m-%d}"
    )
    return end_date, begin_date


class Field(enum.Enum):
    """Common interface for the field names coming from different sources."""

    SYMBOL = "symbol"
    DATE = "date"
    CLOSE = "close"
    TIME = "time"
    TIME_STAMP = "time_stamp"
    OPEN = "open"
    LOW = "low"
    HIGH = "high"
    VOLUME = "volume"
    SOURCE = "source"
    BID = "bid"
    BID_SIZE = "bid_size"
    ASK = "ask"
    ASK_SIZE = "ask_size"
    PCT_CHANGE = "pct_change"
    CHANGE = "change"
    PREVIOUS_CLOSE = "previous_close"
    SPLIT_ADJ = "split_adj"
    DIST_ADJ = "dist_adj"
    EX_DATE = "ex_date"
    PAY_DATE = "pay_date"
    AMOUNT = "amount"
    NAME = "name"
    N_SHARES_OUTSTANDING = "n_shares_outstanding"
    EXCHANGE = "exchange"
    CURRENCY = "currency"


def to_decimal(val: D | str | float) -> D:
    """Convert val to decimal.Decimal."""
    if isinstance(val, D):
        return val
    if isinstance(val, int | float):
        return D(str(val))
    return D(val)


def to_decimal_or_none(val: None | D | str | float) -> D | None:
    """Convert val to decimal.Decimal."""
    if val is None:
        return None
    try:
        return to_decimal(val)
    except (ValueError, TypeError, decimal.InvalidOperation) as err:
        if val is not None:
            log.warning("cannot convert %s to decimal: %s", val, err)
    return None


def to_int_or_none(val: None | D | str | float) -> int | None:
    """Convert val to int."""
    if val is None:
        return None
    try:
        return int(val)
    except (ValueError, TypeError, decimal.InvalidOperation):
        try:
            return int(float(val))
        except (ValueError, TypeError, decimal.InvalidOperation) as err:
            if val is not None:
                log.warning("cannot convert %s to int: %s", val, err)
    return None


@functools.total_ordering
class Price:  # pylint: disable=R0902, R0903
    """Hold price information.

    Use vars(historical_result) to cached_get a dict-like interface.

    All numeric fields are expected to be decimal.Decimal.

    Time is present and separate from date because the data may be known for
    intraday data but it is usually not specified for end of day data.

    """

    # Slots limits memory usage and ensures mispelling are caught.
    __slots__ = [
        "_close",
        "_date_or_datetimetz",
        "_high",
        "_low",
        "_open",
        "_volume",
        "dist_adj",
        "source",
        "split_adj",
        "symbol",
    ]
    # https://en.wikipedia.org/wiki/ISO_8601
    __date_fmt = "%Y-%m-%d"
    __datetime_fmt = "%Y-%m-%dT%H:%M:%S"
    # Length of a date-only string in YYYY-MM-DD format
    _DATE_ONLY_LENGTH = 10

    @classmethod
    def __convert_date_internal(cls, date: dt.date | dt.datetime) -> dt.date | dt.datetime:
        """Convert date or datetime to tz-aware datetime."""
        if isinstance(date, dt.datetime):
            if date.tzinfo is None:
                return date.replace(tzinfo=default_timezone())
            return date
        if isinstance(date, dt.date):
            return date
        msg = f"date or datetime expected, got {type(date)}"
        raise TypeError(msg)

    # pylint: disable=R0913
    def __init__(
        self: t.Self,
        symbol: str,
        date: dt.date | dt.datetime,
        close: D | float | str,
        open_: D | float | str | None = None,
        low: D | float | str | None = None,
        high: D | float | str | None = None,
        volume: D | int | str | None = None,
        source: str | None = None,
    ) -> None:
        """Initialize Price.

        :param symbol: security symbol
        :param date: price observation date, can be a date or a tz-aware datetime.
        :param close: close price (unadjusted)
        :param open_: open price (unadjusted)
        :param high: high price (unadjusted)
        :param low: low price (unadjusted)
        :param volume: volume for that bar (unadjusted)
        :param source: data source

        Note that close may refer to the last or to the close depending on the
        context (e.g. day close, or interval close or last price).

        """
        if not isinstance(symbol, str):
            msg = f"symbol: str expected, got {type(symbol)}"
            raise TypeError(msg)
        self.symbol = symbol
        self._date_or_datetimetz: dt.date | dt.datetime = self.__convert_date_internal(date)
        # close is required, so we need to give it a value even if it is 0.0.
        self._close = to_decimal(close)
        self._open = to_decimal_or_none(open_)
        self._high = to_decimal_or_none(high)
        self._low = to_decimal_or_none(low)
        self._volume = to_int_or_none(volume) if volume is not None else None
        self.source = source

    def __repr__(self: t.Self) -> str:
        """Representation of Price.

        Looks like code needed to instatiate the same object.

        """

        def __quote(val: None | int | D | str) -> str:
            return f"'{val}'" if val is not None else "None"

        return (
            f"{self.__class__.__name__}(symbol={__quote(self.symbol)}, "
            f"date={self.date!r}, close={__quote(self.close)}, "
            f"open_={__quote(self.open)}, "
            f"low={__quote(self.low)}, high={__quote(self.high)}, "
            f"volume={__quote(self.volume)}, source={__quote(self.source)})"
        )

    def __str__(self: t.Self) -> str:
        """Forward to to_csv()."""
        return self.to_csv()

    def to_csv(self: t.Self, date_format: str | None = None) -> str:
        """Conversion to string."""
        date_format = date_format or (self.__datetime_fmt if self.time else self.__date_fmt)
        retval = f"{self.symbol}," if self.symbol else ","
        if self.date:
            if self.time:
                comb = dt.datetime.combine(self.date, self.time)
                retval += f"{dt.datetime.strftime(comb, date_format)},"
            else:
                retval += f"{dt.date.strftime(self.date, date_format)},"
        retval += f"{self.close}," if self.close else ","
        retval += f"{self.open}," if self.open else ","
        retval += f"{self.high}," if self.high else ","
        retval += f"{self.low}," if self.low else ","
        retval += f"{self.volume:.0f}," if self.volume else ","
        retval += f"{self.source}" if self.source else ","
        return retval

    @classmethod
    def get_datetime(cls, datetimestr: str, date_format: str | None = None) -> dt.datetime:
        """Get tz-aware datetime from string.

        :param datetimestr: date string in YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS format
        :param date_format: overrides the Price.__date_fmt or Price.__datetime_fmt
        :return: datetime object with timezone info if available, otherwise date and None for time

        :raises ValueError: if the date string is not in the expected format
        """
        date_format = date_format or (
            Price.__datetime_fmt if len(datetimestr) > Price._DATE_ONLY_LENGTH else Price.__date_fmt
        )
        try:
            dt_obj = dt.datetime.strptime(datetimestr, date_format).replace(tzinfo=default_timezone())
        except ValueError as error:
            msg = f"invalid date string: {datetimestr} with format {date_format}"
            raise ValueError(msg) from error
        else:
            return dt_obj

    @classmethod
    def from_csv(
        cls,
        price_str: str,
        date_format: str | None = None,
    ) -> "Price":  # pylint: disable=too-many-locals
        """Convert csv string to finquotes.Price.

        This function reverse str(). If no the format is a date, the time is set
        to None to avoid tz issues.

        :param price_str: pice string defined as:
            symbol, date (YYYY-MM-DD), close, open, high, low, volume, source
        :param date_format: overrides the Price.__date_fmt or Price.__datetime_fmt

        """
        symbol, date_str, close, open_, high, low, volume, *other = price_str.split(",")
        date_str = date_str.strip()
        dt_obj = cls.get_datetime(date_str, date_format)
        date = dt_obj.date()
        time = None
        if len(date_str) > cls._DATE_ONLY_LENGTH:
            # If the date string is longer than the date format, it contains time.
            time = dt_obj.timetz()
        source = None
        if other:
            source = other[0]
        price = Price(symbol, date, D(close), source=source)
        price.open = D(open_) if open_ else None
        price.high = D(high) if high else None
        price.low = D(low) if low else None
        price.volume = int(volume) if volume else None
        price.time = time
        return price

    def __lt__(self: t.Self, other: t.Any) -> bool:
        """Compare based on symbol, then date, time, and close."""
        if not isinstance(self, type(other)):
            return NotImplemented
        return (self.symbol, self.date, self.time) < (
            other.symbol,
            other.date,
            other.time,
        )

    def __eq__(self: t.Self, other: object) -> bool:
        """Equal if every field is the same (including symbol)."""
        if not isinstance(other, Price):
            return False
        return (
            self.symbol == other.symbol
            and self.close == other.close
            and self.date == other.date
            and self._open == other._open
            and self._high == other._high
            and self._low == other._low
            and self._volume == other._volume
            and self.source == other.source
        )

    def __hash__(self: t.Self) -> int:
        """Hash based on all fields used in equality comparison."""
        return hash(
            (
                self.symbol,
                self.close,
                self.date,
                self._open,
                self._high,
                self._low,
                self._volume,
                self.source,
            ),
        )

    @property
    def datetime(self: t.Self) -> dt.datetime | None:
        """Get tz-aware datetime for price."""
        if isinstance(self._date_or_datetimetz, dt.datetime):
            return self._date_or_datetimetz
        assert self._date_or_datetimetz is not None
        return dt.datetime.combine(self._date_or_datetimetz, dt.time(0, 0), tzinfo=default_timezone())

    @datetime.setter
    def datetime(self: t.Self, value: dt.datetime) -> None:
        """Set the datetime of the price.

        :param value: datetime with timezone info.

        """
        self._date_or_datetimetz = self.__convert_date_internal(value)

    @property
    def date(self: t.Self) -> dt.date:
        """Get date for price."""
        if isinstance(self._date_or_datetimetz, dt.datetime):
            return self._date_or_datetimetz.date()
        assert self._date_or_datetimetz is not None
        return self._date_or_datetimetz

    @date.setter
    def date(self: t.Self, value: dt.date | dt.datetime) -> None:
        """Set the date of the price.

        :param value: date.

        Accept a tz-aware datetime as well and forwards the call to the .datetime property in that case.

        """
        self._date_or_datetimetz = self.__convert_date_internal(value)

    @property
    def time(self: t.Self) -> dt.time | None:
        """Get tz-aware time of price."""
        if isinstance(self._date_or_datetimetz, dt.datetime):
            return self._date_or_datetimetz.timetz()
        return None

    @time.setter
    def time(self: t.Self, value: dt.time | None) -> None:
        """Set time of price.

        :param value: time with timezone info or None

        If the time is not tz-aware, it will be combined with the date and follow
        the logic of .date assignment.

        :raises AssertionError: if value is not None and not a datetime.time

        """
        if value is None:
            self._date_or_datetimetz = self.date
            return
        assert isinstance(value, dt.time), f"time: time expected, got {type(value)}"
        self.date = dt.datetime.combine(self.date, value)

    @property
    def close(self: t.Self) -> D:
        """Get the close or last depending on context."""
        assert self._close is not None  # for mypy
        return self._close

    @close.setter
    def close(self: t.Self, value: D | str | float) -> None:
        """Set the price or last depending on context."""
        self._close = to_decimal(value)

    @property
    def open(self: t.Self) -> D | None:
        """Get the price interval open."""
        return self._open

    @open.setter
    def open(self: t.Self, value: D | None) -> None:
        """Set the price interval open."""
        assert value is None or isinstance(value, D)
        self._open = value

    @property
    def low(self: t.Self) -> D | None:
        """Get the price interval low."""
        return self._low

    @low.setter
    def low(self: t.Self, value: D | None) -> None:
        """Set the price interval low."""
        assert value is None or isinstance(value, D)
        self._low = value

    @property
    def high(self: t.Self) -> D | None:
        """Get the price interval high."""
        return self._high

    @high.setter
    def high(self: t.Self, value: D | None) -> None:
        """Set the price interval high."""
        assert value is None or isinstance(value, D)
        self._high = value

    @property
    def volume(self: t.Self) -> int | None:
        """Get the interval volume."""
        return self._volume

    @volume.setter
    def volume(self: t.Self, value: int | None) -> None:
        """Set the interval volume."""
        assert value is None or isinstance(value, int)
        self._volume = value


class PriceValidationDiagnosticLevel(enum.Enum):
    """Validation level."""

    OK = enum.auto()
    WARNING = enum.auto()
    ERROR = enum.auto()


@dataclasses.dataclass
class PriceValidationDiagnostic:
    """Validation status."""

    level: PriceValidationDiagnosticLevel
    message: str

    def __init__(
        self: t.Self,
        level: PriceValidationDiagnosticLevel = PriceValidationDiagnosticLevel.OK,
        message: str = "",
    ) -> None:
        """Initialize."""
        self.message = message
        self.level = level


def check_relative_attribute_values(
    price: t.Any,
    tol: D = D("0.01"),
) -> PriceValidationDiagnostic:
    """Check that the relative attributes are consistent."""
    if price.low is not None:
        if price.close and price.close < price.low - tol:
            return PriceValidationDiagnostic(
                PriceValidationDiagnosticLevel.WARNING,
                f"close {price.close:.2f} lower than low {price.low:.2f}",
            )
        if price.open and price.open < price.low - tol:
            return PriceValidationDiagnostic(
                PriceValidationDiagnosticLevel.WARNING,
                f"open {price.open:.2f} lower than low {price.low:.2f}",
            )
    if price.high is not None:
        if price.close and price.close > price.high + tol:
            return PriceValidationDiagnostic(
                PriceValidationDiagnosticLevel.WARNING,
                f"close {price.close:.2f} higher than high {price.high:.2f}",
            )
        if price.open and price.open > price.high + tol:
            return PriceValidationDiagnostic(
                PriceValidationDiagnosticLevel.WARNING,
                f"open {price.open:.2f} higher than high {price.high:.2f}",
            )
    return PriceValidationDiagnostic(PriceValidationDiagnosticLevel.OK, "")


def check_date_too_old(price: Price) -> PriceValidationDiagnostic:
    """Check that the date is not too old."""
    if price.date is not None and price.date < dt.date(1900, 1, 1):
        return PriceValidationDiagnostic(
            PriceValidationDiagnosticLevel.ERROR,
            f"date {price.date} is earlier than 1900-01-01",
        )
    return PriceValidationDiagnostic(PriceValidationDiagnosticLevel.OK, "")


def _make_default_checks() -> list[t.Callable[[t.Any], PriceValidationDiagnostic]]:
    """Make default checks."""
    checks: list[t.Callable[[t.Any], PriceValidationDiagnostic]] = []

    def _create_missing_attribute_check(
        attribute: str,
        level: PriceValidationDiagnosticLevel,
    ) -> t.Callable[[t.Any], PriceValidationDiagnostic]:
        """Create a function to check for missing attributes."""

        def check_missing_attribute(price: t.Any) -> PriceValidationDiagnostic:
            return (
                PriceValidationDiagnostic(level, f"missing {attribute}")
                if not getattr(price, attribute)
                else PriceValidationDiagnostic()
            )

        return check_missing_attribute

    for attribute, level in [
        ("symbol", PriceValidationDiagnosticLevel.ERROR),
        ("date", PriceValidationDiagnosticLevel.ERROR),
        ("close", PriceValidationDiagnosticLevel.ERROR),
        ("source", PriceValidationDiagnosticLevel.WARNING),
        ("time", PriceValidationDiagnosticLevel.WARNING),
        ("open", PriceValidationDiagnosticLevel.WARNING),
        ("low", PriceValidationDiagnosticLevel.WARNING),
        ("high", PriceValidationDiagnosticLevel.WARNING),
        ("volume", PriceValidationDiagnosticLevel.WARNING),
    ]:
        checks.append(_create_missing_attribute_check(attribute, level))

    def _wrapped_relative_check(price: t.Any) -> PriceValidationDiagnostic:
        return check_relative_attribute_values(price)

    def _wrapped_date_check(price: t.Any) -> PriceValidationDiagnostic:
        return check_date_too_old(price)

    checks.append(_wrapped_relative_check)
    checks.append(_wrapped_date_check)
    return checks


class PriceValidator:
    """Validate price data."""

    _default_checks = _make_default_checks()

    def __init__(
        self: t.Self,
        checks: list[t.Callable[..., PriceValidationDiagnostic]] | None = None,
        as_of_date: dt.date | None = None,
    ) -> None:
        """Initialize the validator."""
        self.checks = checks or self._default_checks
        self.as_of_date = as_of_date or today()
        self.checks.append(self.check_date_is_in_past)

    def check_date_is_in_past(self: t.Self, price: Price) -> PriceValidationDiagnostic:
        """Check that the date is in the past."""
        if price.date is not None and price.date > self.as_of_date:
            return PriceValidationDiagnostic(
                PriceValidationDiagnosticLevel.ERROR,
                f"date {price.date} is in the future",
            )
        return PriceValidationDiagnostic()

    def detect_issues(
        self: t.Self,
        price: Price,
    ) -> t.Iterable[PriceValidationDiagnostic]:
        """Loops on checks and yield all detected issues."""
        for check in self.checks:
            result = check(price)
            if result and result.level != PriceValidationDiagnosticLevel.OK:
                yield result


@functools.total_ordering
class Distribution:  # pylint: disable=R0903
    """Hold a historical dividend record.

    Note: use vars(Distribution) to cached_get a dict-like interface.

    """

    # Slots limits memory usage and ensures mispelling are caught.
    __slots__ = ["_amount", "_ex_date", "_pay_date", "symbol"]

    def __init__(
        self: t.Self,
        symbol: str,
        ex_date: dt.date,
        amount: D,
        pay_date: dt.date | None = None,
    ) -> None:
        """Create new distribution result.

        :param str symbol: security identifier
        :param datetime.date ex_date: ex-date of the distribution
        :param decimal.Decimal amount: amount paid
        :param pay_date: pay date
        :type pay_date: None or datetime.date

        """
        self.symbol = symbol
        self._ex_date = ex_date
        self._amount = to_decimal_or_none(amount)
        self._pay_date = pay_date

    @property
    def ex_date(self: t.Self) -> dt.date:
        """Get ex-date of distribution."""
        return self._ex_date

    @ex_date.setter
    def ex_date(self: t.Self, value: dt.date) -> None:
        """Set ex-date of distribution."""
        assert isinstance(value, dt.date)
        self._ex_date = value

    @property
    def amount(self: t.Self) -> D | None:
        """Get amount of distribution."""
        return self._amount

    @amount.setter
    def amount(self: t.Self, value: D) -> None:
        """Set amount of distribution."""
        assert isinstance(value, D)
        self._amount = value

    @property
    def pay_date(self: t.Self) -> dt.date | None:
        """Get pay date of distribution."""
        return self._pay_date

    @pay_date.setter
    def pay_date(self: t.Self, value: dt.date) -> None:
        """Set pay date of distribution."""
        assert isinstance(value, dt.date)
        self._pay_date = value

    def __repr__(self: t.Self) -> str:
        """Representation of object."""
        return f"<{self.__class__.__name__} {self.symbol} {self.ex_date} {self.amount}>"

    def __str__(self: t.Self) -> str:
        """Representation of object."""
        return f"{self.symbol},{self.ex_date},{self.amount}"

    def __lt__(self: t.Self, other: object) -> bool:
        """Compare for ordering based on symbol, ex_date, then amount."""
        if not isinstance(other, Distribution):
            return NotImplemented
        return (self.symbol, self.ex_date, self.amount) < (
            other.symbol,
            other.ex_date,
            other.amount,
        )

    def __eq__(self: t.Self, other: object) -> bool:
        """Compare symbol, ex-date and amount."""
        if not isinstance(other, Distribution):
            return False
        return self.symbol == other.symbol and self.ex_date == other.ex_date and self.amount == other.amount

    def __hash__(self: t.Self) -> int:
        """Hash based on symbol, ex_date, and amount."""
        return hash((self.symbol, self.ex_date, self.amount))


@functools.total_ordering
class Split:  # pylint: disable=R0903
    """Hold a historical dividend record.

    Note: use vars(Split) to cached_get a dict-like interface.

    """

    # Slots limits memory usage and ensures mispelling are caught.
    __slots__ = ["_ex_date", "_new_quantity", "_old_quantity", "symbol"]

    def __init__(
        self: t.Self,
        symbol: str,
        ex_date: dt.date,
        new_quantity: int,
        old_quantity: int,
    ) -> None:
        """Create new distribution result.

        :param symbol: security symbol
        :param ex_date: ex-date of the split
        :param new_quantity: number of shares after split
        :param old_quantity: number of shares before split

        """
        assert isinstance(new_quantity, int)
        assert isinstance(old_quantity, int)
        assert isinstance(ex_date, dt.date)
        self.symbol = symbol
        self.ex_date = ex_date
        self.new_quantity = new_quantity
        self.old_quantity = old_quantity

    @staticmethod
    def from_ratio(symbol: str, ex_date: dt.date, ratio: D) -> "Split":
        """Create new distribution result.

        :param symbol: security symbol
        :param ex_date: ex-date of the split
        :param ratio: ratio of the split: >1 for split, <1 for reverse split

        """
        assert isinstance(ratio, D)
        assert isinstance(ex_date, dt.date)
        new_quantity, old_quantity = ratio.as_integer_ratio()
        reverse_ratio = (1 / ratio).quantize(D("0.0001"))
        reverse_old_quantity, reverse_new_quantity = reverse_ratio.as_integer_ratio()
        if new_quantity + old_quantity < reverse_new_quantity + reverse_old_quantity:
            return Split(
                symbol=symbol,
                ex_date=ex_date,
                new_quantity=int(new_quantity),
                old_quantity=int(old_quantity),
            )
        return Split(
            symbol=symbol,
            ex_date=ex_date,
            new_quantity=int(reverse_new_quantity),
            old_quantity=int(reverse_old_quantity),
        )

    @property
    def ex_date(self: t.Self) -> dt.date:
        """Get ex-date of distribution."""
        return self._ex_date

    @ex_date.setter
    def ex_date(self: t.Self, value: dt.date) -> None:
        """Set ex-date of distribution."""
        assert isinstance(value, dt.date)
        self._ex_date = value

    @property
    def new_quantity(self: t.Self) -> int:
        """Get new shares quantity."""
        return self._new_quantity

    @new_quantity.setter
    def new_quantity(self: t.Self, value: int) -> None:
        """Set new shares quantity."""
        assert isinstance(value, int)
        self._new_quantity = value

    @property
    def old_quantity(self: t.Self) -> int:
        """Get old shares quantity."""
        return self._old_quantity

    @old_quantity.setter
    def old_quantity(self: t.Self, value: int) -> None:
        """Set old shares quantity."""
        assert isinstance(value, int)
        self._old_quantity = value

    def __repr__(self: t.Self) -> str:
        """Representation of object."""
        return f"<{self.__class__.__name__} {self.symbol} {self.ex_date} {self.new_quantity}/{self.old_quantity}>"

    def __str__(self: t.Self) -> str:
        """Representation of object."""
        return f"{self.symbol},{self.ex_date},{self.new_quantity}/{self.old_quantity}"

    def __lt__(self: t.Self, other: object) -> bool:
        """Compare for ordering based on symbol, ex_date, then ratio."""
        if not isinstance(other, Split):
            return NotImplemented
        return (self.symbol, self.ex_date, self.new_quantity, self.old_quantity) < (
            other.symbol,
            other.ex_date,
            other.new_quantity,
            other.old_quantity,
        )

    def __eq__(self: t.Self, other: object) -> bool:
        """Compare symbol, ex-date and new and old quantity."""
        if not isinstance(other, Split):
            return False
        return (
            self.symbol == other.symbol
            and self.ex_date == other.ex_date
            and self.new_quantity == other.new_quantity
            and self.old_quantity == other.old_quantity
        )

    def __hash__(self: t.Self) -> int:
        """Hash based on symbol, ex_date, new_quantity, and old_quantity."""
        return hash((self.symbol, self.ex_date, self.new_quantity, self.old_quantity))


@dataclasses.dataclass(frozen=True)
class _ExchangeData:
    iso_country: str
    mic: str
    operating_mic: str
    acronym: str
    city: str


def _get_exchanges() -> dict[str, _ExchangeData]:
    """Get a list of valid exchanges.

    See https://www.iso20022.org/sites/default/files/ISO10383_MIC/ISO10383_MIC.csv

    """

    def __parse_io(fh: t.Iterable[str]) -> dict[str, _ExchangeData]:
        reader = csv.reader(fh)
        next(reader)  # skip header
        return {
            row[2]: _ExchangeData(
                iso_country=row[1],
                mic=row[2],
                operating_mic=row[3],
                acronym=row[6],
                city=row[7],
            )
            for row in reader
        }

    def parse_csv(
        file_or_buffer: importlib.resources.abc.Traversable | str | io.TextIOBase,
        encoding: str | None = None,
    ) -> dict[str, _ExchangeData]:
        """Parse a csv file or buffer and return a dict of exchanges.

        :param file_or_buffer: path to csv file, buffer containing the csv data, or file-like object.
        :param encoding: encoding of the csv file.
        :return: dict of exchanges.
        """
        # Check if file_or_buffer is a string that contains CSV data
        if isinstance(file_or_buffer, str):
            str_data: str = file_or_buffer
            if str_data.startswith('"COUNTRY"'):
                return __parse_io(io.StringIO(file_or_buffer))
            with Path(str_data).open(encoding=encoding) as fh:
                return __parse_io(fh)
        # Check if file_or_buffer is a Traversable object (e.g., from importlib.resources)
        if isinstance(file_or_buffer, importlib.resources.abc.Traversable):
            with file_or_buffer.open("r", encoding=encoding) as fh:
                return __parse_io(fh)
        # Assume file_or_buffer is a file-like object (e.g., io.StringIO or io.BytesIO)
        if isinstance(file_or_buffer, io.IOBase):
            return __parse_io(file_or_buffer)
        msg = "Unsupported type for file_or_buffer: must be a file path, buffer, or file-like object."
        raise ValueError(
            msg,
        )

    url = "https://www.iso20022.org/sites/default/files/ISO10383_MIC/ISO10383_MIC.csv"
    try:
        data = localhttp.request(localhttp.cached_session, "GET", url).text
        exchanges = parse_csv(data)
    except (
        OSError,
        csv.Error,
        AttributeError,
        urllib3.exceptions.HTTPError,
    ) as error:
        log.info("failed to fetch %s: %s. Falling back to embeded file", url, error)
        iso10383 = importlib.resources.files(__name__) / "data/ISO10383_MIC.csv"
        assert iso10383.is_file(), str(iso10383) + " file not found"
        exchanges = parse_csv(iso10383, encoding="utf-8")
    return exchanges


@functools.cache
def get_exchanges() -> dict[str, _ExchangeData]:
    """Cache layer for _get_exchanges()."""
    return _get_exchanges()


@dataclasses.dataclass(frozen=True)
class Exchange:
    """Hold an exchange record.

    The mic value validity is checked only in the constructor so you will need to change that
    check if you want to change the mic.  Seems more efficient to require to create new instances
    of this class instead of modifying its only field.

    :seealso: get_operating_exchange()

    """

    mic: str

    def __post_init__(self: t.Self) -> None:
        """Check mic value."""
        if self.mic is None:
            msg = "mic is required"
            raise ValueError(msg)
        if self.mic not in get_exchanges():
            msg = f"invalid mic: {self.mic}"
            raise ValueError(msg)

    @property
    def name(self: t.Self) -> str:
        """Get name of exchange."""
        return get_exchanges()[self.mic].acronym

    @property
    def city(self: t.Self) -> str:
        """Get country of exchange."""
        return get_exchanges()[self.mic].city

    @property
    def operating_mic(self: t.Self) -> str:
        """Get operating MIC."""
        return get_exchanges()[self.mic].operating_mic

    def __str__(self: t.Self) -> str:
        """MIC code for the exchange."""
        return f"{self.mic}"


def get_operating_exchange(exchange: Exchange) -> Exchange:
    """Consolidate exchange record to the operating exchange."""
    return Exchange(mic=exchange.operating_mic)


@dataclasses.dataclass
class Security:
    """Hold security information."""

    symbol: str
    name: str
    type: str | None = None
    currency: str | None = None
    exchange: Exchange | None = None
    country: str | None = None
    sector: str | None = None
    industry: str | None = None
    source: str | None = None
    # If the source has an identifier for that security, slot it here.
    source_id: str | None = None

    def __post_init__(self: t.Self) -> None:
        """Set attribute to lower case and run checks."""
        self._assert_invariants()

    def _assert_invariants(self: t.Self) -> None:
        """Assert invariants."""
        assert self.source is not None, "source is None!"

    def __str__(self: t.Self) -> str:
        """Representation of object as csv string."""
        return to_csv(self)


@functools.total_ordering
class ShortInterest:  # pylint: disable=too-few-public-methods
    """Hold short interest."""

    # Slots limits memory usage and ensures mispelling are caught.
    __slots__ = ["as_of_date", "n_shares", "name", "symbol"]

    def __init__(
        self: t.Self,
        symbol: str,
        as_of_date: dt.date,
        n_shares: D,
        *args: t.Sequence[t.Any],
        **kwargs: t.Any,
    ) -> None:
        """Create new short interest information.

        :param symbol: security ticker.
        :param as_of_date: date for which short interest was measured.
        :param n_shares: numbed of shares sold short.
        :param args: optional parameters to set attributes.

        """
        assert isinstance(symbol, str)
        self.symbol = symbol
        self.n_shares = n_shares
        assert isinstance(as_of_date, dt.date)
        self.as_of_date = as_of_date
        for akey, avalue in zip(self.__slots__, args, strict=False):
            setattr(self, akey, avalue)
        for kwkey, kwvalue in kwargs.items():
            setattr(self, kwkey, kwvalue)

    def __str__(self: t.Self) -> str:
        """Output of str() function.

        :rtype: str.

        """
        return f"{self.symbol} {self.as_of_date} {self.n_shares}"

    def __repr__(self: t.Self) -> str:
        """Output of repr() function.

        :rtype: str.

        """
        return "<{self.__class__.__name__} symbol:{self.symbol} as_of_date:{self.as_of_date} n_shares:{self.n_shares}>"

    def __eq__(self: t.Self, other: object) -> bool:
        """Check equality with another ShortInterest object."""
        if not isinstance(other, ShortInterest):
            return False
        return self.symbol == other.symbol and self.as_of_date == other.as_of_date

    def __hash__(self: t.Self) -> int:
        """Hash based on symbol and as_of_date."""
        return hash((self.symbol, self.as_of_date))

    def __lt__(self: t.Self, other: object) -> bool:
        """Compare for ordering based on symbol, as_of_date, then n_shares."""
        if not isinstance(other, ShortInterest):
            return NotImplemented
        return (self.symbol, self.as_of_date, self.n_shares) < (
            other.symbol,
            other.as_of_date,
            other.n_shares,
        )


class AbstractFeed(metaclass=abc.ABCMeta):  # pylint: disable=R0903
    """Factor common attributes to all finquotes.Feed.

    Actual abstract method will be defined in the subclasses of AbstractFeed.

    .. seealso::
        AbstractCloseFeed, AbstractHistoricalFeed, AbstractDistributionFeed,
        AbstractSplitFeed.

    """

    def __init__(
        self: t.Self,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the close class.

        :param name: identify the feed (e.g. finquotes.yahoo).
        :param source: identify the data source (e.g. yahoo).

        """
        self.name = kwargs.pop("name")
        self.source = kwargs.pop("source")

        # Call super() in case there are other classes in the MRO
        # This is safe even if object.__init__ is next (it ignores kwargs)
        super().__init__()

    @property
    @abc.abstractmethod
    def timeout(self: t.Self) -> float:
        """Get timeout in seconds for network operations."""
        msg = "Implement timeout property or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @timeout.setter
    @abc.abstractmethod
    def timeout(self: t.Self, value: float) -> None:
        """Set timeout in seconds for network operations."""
        msg = "Implement timeout setter or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @property
    @abc.abstractmethod
    def max_retries(self: t.Self) -> int:
        """Get maximum number of retries for network operations."""
        msg = "Implement max_retries property or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @max_retries.setter
    @abc.abstractmethod
    def max_retries(self: t.Self, value: int) -> None:
        """Set maximum number of retries for network operations."""
        msg = "Implement max_retries setter or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @property
    @abc.abstractmethod
    def retry_delay(self: t.Self) -> float:
        """Get delay between retries in seconds."""
        msg = "Implement retry_delay property or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @retry_delay.setter
    @abc.abstractmethod
    def retry_delay(self: t.Self, value: float) -> None:
        """Set delay between retries in seconds."""
        msg = "Implement retry_delay setter or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @property
    @abc.abstractmethod
    def api_key(self: t.Self) -> str | None:
        """Get API key for authentication."""
        msg = "Implement api_key property or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)

    @api_key.setter
    @abc.abstractmethod
    def api_key(self: t.Self, value: str | None) -> None:
        """Set API key for authentication."""
        msg = "Implement api_key setter or inherit from NetworkConfigMixin for default implementation"
        raise NotImplementedError(msg)


class AbstractCloseFeed(AbstractFeed):
    """Interface for drivers implementing retrieval of close prices.

    Note those closes may be (and often are) available after the close only.
    They still qualify as close prices.
    """

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        as_of_date: dt.date | None = None,
        batch_size: int | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the close class.

        :param name: passed to AbstractFeed.
        :param source: passed to AbstractFeed.
        :param as_of_date: as of date being looked for.
        :param batch_size: number of securities in each request.
        :param kwargs: Additional arguments passed to parent classes.

        """
        self.as_of_date = as_of_date or today()
        self.batch_size = batch_size or 100
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
        **kwargs: t.Any,
    ) -> t.Generator[Price, None, None]:
        """Retrieve current prices.

        :param symbols: symbols to looked up.
        :param date: date for which to look up prices (defaults to as_of_date).

        Yield a generator iterating over Price.

        """
        raise NotImplementedError


class AbstractIntraDayFeed(AbstractFeed):
    """Interface for drivers implementing retrieval of intraday prices."""

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_intraday_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        as_of_ts: dt.datetime | None = None,
        fields: t.Sequence[Field] | None = None,
    ) -> t.Generator[Price, None, None]:
        """Retrieve current prices.

        :param symbols: symbols looked up
        :param as_of_ts: expected as_of_date
        :param fields: optional list of fields to retrieve

        :return: generator of Price objects
        """
        raise NotImplementedError


class AbstractHistoricalFeed(AbstractFeed):  # pylint: disable=R0903
    """Interface of historical price feed.

    Derived class must implement:

    def fetch_hist_prices(
        self,
        symbol,
        end_date=None,
        begin_date=None,
        frequency=None,
        splits=None,
        distributions=None,
    )
    """

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    # pylint: disable=R0913
    @abc.abstractmethod
    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        splits: t.Sequence[Split] | None = None,
        distributions: t.Sequence[Distribution] | None = None,
    ) -> t.Generator[Price, None, None]:
        """Retrieve price history.

        :return: generator of Price objects

        """
        raise NotImplementedError


class AbstractDistributionFeed(AbstractFeed):
    """Interface of distribution feeds."""

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_hist_dists(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[Distribution, None, None]:
        """Retrieve distribution history (dividend, capital gains, ...).

        :param symbol: Security symbol
        :param end_date: End date for the period to retrieve (inclusive)
        :param begin_date: Start date for the period to retrieve (inclusive)
        :return: Generator of Distribution objects
        """
        raise NotImplementedError


class AbstractSplitFeed(AbstractFeed):
    """Interface of distribution feeds."""

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_hist_splits(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[Split, None, None]:
        """Retrieve split history.

        :param symbol: Security symbol
        :param end_date: End date for the period to retrieve (inclusive)
        :param begin_date: Start date for the period to retrieve (inclusive)

        :return: generator of Split objects
        """
        raise NotImplementedError


class AbstractShortInterestFeed(AbstractFeed):
    """Interface of short interest feeds."""

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_shortints(
        self: t.Self,
        symnexchs: list[tuple[str, str, str]],
        as_of_date: dt.date,
    ) -> t.Generator[ShortInterest, None, None]:
        """Retrieve short interests.

        :param symnexchs: List of (symbol, exchange, exchange_code) triplets to identify stocks
        :param as_of_date: Date to retrieve short interest for

        :return: Generator of ShortInterest objects
        """
        raise NotImplementedError


class AbstractSecurityFeed(AbstractFeed):
    """Interface to security feed.

    Retrieve security information with this feed.

    """

    def __init__(
        self: t.Self,
        name: str,
        source: str,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed.

        :param kwargs: Additional arguments passed to parent classes.
        """
        super().__init__(name=name, source=source, **kwargs)

    @abc.abstractmethod
    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[Security, None, None]:
        """Retrieve security information.

        :param symbols: Symbols to retrieve information for
        :return: Generator of Security objects
        """
        raise NotImplementedError


class FeedType(enum.Enum):
    """Identify feed type when calling get_feed()."""

    CLOSE = enum.auto()
    INTRADAY = enum.auto()
    HISTORICAL = enum.auto()
    DISTRIBUTION = enum.auto()
    SPLIT = enum.auto()
    SECURITY = enum.auto()
    SHORT = enum.auto()

    def __str__(self: t.Self) -> str:
        """Return simpe string representation so it plays nice with pytest (no dot, no FeedType)."""
        return self.name.lower()


def build_feed(
    feed_name: str,
    feed_type: FeedType,
    **kwargs: t.Any | None,
) -> AbstractFeed:
    """Create and return a new instance of a feed class.

    Consider using build_* functions instead because they return a typed feed.

    :param feed_name: module name with entry point(s) to build concrete feeds
    :param feed_type: type of feed
    :param kwargs: additional parameters passed to the concrete feed constructor

    :return: An instance of the requested feed type
    :raises ImportError: if the module cannot be imported
    :raises ValueError: if no valid feed can be created
    :raises AttributeError: if the module doesn't implement the requested feed builder
    """
    if not isinstance(feed_type, FeedType):
        msg = f"feed_type must be a FeedType enum, got {type(feed_type)}"
        raise TypeError(msg)
    if not isinstance(feed_name, str):
        msg = f"feed_name must be a string, got {type(feed_name)}"
        raise TypeError(msg)

    log.info("loading %s %s ...", feed_name, feed_type)
    module: t.Any = None
    try:
        module = importlib.import_module(feed_name)
    except ImportError:
        log.exception("cannot import %s", feed_name)
        raise

    builder_map = {
        FeedType.CLOSE: "build_close_feed",
        FeedType.INTRADAY: "build_intraday_feed",
        FeedType.HISTORICAL: "build_historical_feed",
        FeedType.DISTRIBUTION: "build_distribution_feed",
        FeedType.SPLIT: "build_split_feed",
        FeedType.SECURITY: "build_security_feed",
        FeedType.SHORT: "build_short_interest_feed",
    }

    builder_name = builder_map[feed_type]
    try:
        builder = getattr(module, builder_name)
        feed = builder(**kwargs)
    except AttributeError:
        log.exception("Module %s doesn't implement %s", feed_name, builder_name)
        raise

    if not feed:
        msg = f"failed to retrieve a valid {feed_type} feed in {feed_name}"
        raise ValueError(msg)

    return feed


def build_close_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractCloseFeed:
    """Forward call to get_feed() for and cast result to AbstractCloseFeed."""
    return t.cast(
        "AbstractCloseFeed",
        build_feed(
            feed_name,
            FeedType.CLOSE,
            **kwargs,
        ),
    )


def build_intraday_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractIntraDayFeed:
    """Forward call to get_feed() for and cast result to AbstractIntraDayFeed."""
    return t.cast(
        "AbstractIntraDayFeed",
        build_feed(
            feed_name,
            FeedType.INTRADAY,
            **kwargs,
        ),
    )


def build_historical_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractHistoricalFeed:
    """Forward call to get_feed() for and cast result to AbstractHistoricalFeed."""
    return t.cast(
        "AbstractHistoricalFeed",
        build_feed(
            feed_name,
            FeedType.HISTORICAL,
            **kwargs,
        ),
    )


def build_distribution_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractDistributionFeed:
    """Forward call to get_feed() for and cast result to AbstractDistributionFeed."""
    return t.cast(
        "AbstractDistributionFeed",
        build_feed(
            feed_name,
            FeedType.DISTRIBUTION,
            **kwargs,
        ),
    )


def build_split_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractSplitFeed:
    """Forward call to get_feed() for and cast result to AbstractSplitFeed."""
    return t.cast(
        "AbstractSplitFeed",
        build_feed(
            feed_name,
            FeedType.SPLIT,
            **kwargs,
        ),
    )


def build_security_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractSecurityFeed:
    """Forward call to get_feed() for and cast result to AbstractSecurityFeed."""
    return t.cast(
        "AbstractSecurityFeed",
        build_feed(
            feed_name,
            FeedType.SECURITY,
            **kwargs,
        ),
    )


def build_short_interest_feed(
    feed_name: str,
    **kwargs: t.Any,
) -> AbstractShortInterestFeed:
    """Forward call to get_feed() for and cast result to AbstractShortInterestFeed."""
    return t.cast(
        "AbstractShortInterestFeed",
        build_feed(
            feed_name,
            FeedType.SHORT,
            **kwargs,
        ),
    )


def to_csv(dataclass_obj: t.Any) -> str:
    """Return a CSV representation of a dataclass object.

    :param dataclass_obj: An instance of a dataclass
    :return: A CSV string with all dataclass fields
    :raises TypeError: If the object is not a dataclass
    """
    if not dataclasses.is_dataclass(dataclass_obj):
        msg = f"Expected a dataclass instance, got {type(dataclass_obj)}"
        raise TypeError(msg)

    return ",".join(str(getattr(dataclass_obj, field.name) or "") for field in dataclasses.fields(dataclass_obj))


class FeedError(RuntimeError):
    """Error related to a feed behaviour (E.g. not available)."""

    def __init__(self: t.Self, msg: str) -> None:
        super().__init__(msg)



"""Interface to nasdaq.com."""

import dataclasses
import typing
import typing as t

import finquotes as fq

SOURCE = "finquotes.nasdaq"


@dataclasses.dataclass(frozen=True)
class Stock:
    """Listed stock data."""

    symbol: str
    name: str
    country: str
    industry: str
    sector: str

    @classmethod
    def from_dict(cls, json_data: dict[str, typing.Any]) -> "Stock":
        """Convert a dict (could be coming from JSON data) to a Stock class."""
        return cls(
            symbol=json_data["symbol"],
            name=json_data["name"].replace(" Common Stock", "").strip(),
            country=json_data["country"],
            industry=json_data["industry"],
            sector=json_data["sector"],
        )


def _read_json(
    url: str,
    headers: dict | None = None,
    params: t.Any = None,
) -> dict[str, Stock]:
    """Download nasdaq screener data (csv url) or directly json data."""
    json = fq.localhttp.request(
        fq.localhttp.cached_session,
        "GET",
        url,
        headers=headers,
        params=params,
    ).json()
    return {stock["symbol"]: Stock.from_dict(stock) for stock in json["data"]["rows"]}


class NasdaqSecurityFeed(fq.NetworkConfigMixin, fq.AbstractSecurityFeed):
    """Interface to https://api.nasdaq.com.

    From https://github.com/ranaroussi/yfinance/issues/606/ for details.

    """

    __default_url = "https://api.nasdaq.com/api/screener/stocks"
    __default_headers: t.ClassVar[dict[str, str]] = {
        "authority": "api.nasdaq.com",
        "accept": "application/json, text/plain, */*",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)",
        "origin": "https://www.nasdaq.com",
        "sec-fetch-site": "same-site",
        "sec-fetch-mode": "cors",
        "sec-fetch-dest": "empty",
        "referer": "https://www.nasdaq.com/",
        "accept-language": "en-US,en;q=0.9",
    }
    __default_params = params = (
        ("tableonly", "true"),
        ("limit", "25"),
        ("offset", "0"),
        ("download", "true"),
    )

    def __init__(
        self: t.Self,
        url: str | None = None,
        headers: dict | None = None,
        params: t.Any = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize."""
        super().__init__(name=__name__, source=SOURCE, **kwargs)
        self.url = url or self.__default_url
        self.headers = headers or self.__default_headers
        self.params = params or self.__default_params
        self.__stocks: dict[str, Stock] = {}

    @property
    def stocks(self: t.Self) -> dict[str, Stock]:
        """Get nasdaq screener csv."""
        if not self.__stocks:
            assert self.url is not None
            self.__stocks = _read_json(self.url, self.headers, self.params)
        assert self.__stocks is not None
        return self.__stocks

    def approximate_shares_outstanding(
        self: t.Self,
        df: t.Any,
        symbol: str,
    ) -> float | None:
        """Approximate shares outstanding from market cap and price."""
        # Approximation
        try:
            shares_outstanding = df["market_cap"].astype(float) / df["price"].astype(
                float,
            )
        except ValueError as err:
            fq.log.warning(
                "Could not calculate shares outstanding for %s: %s",
                symbol,
                err,
            )
            shares_outstanding = None
        return shares_outstanding

    def fetch_security(self: t.Self, symbol: str) -> fq.Security:
        """Import stock data from https://www.nasdaq.com/market-activity/stocks/screener."""
        stock = self.stocks[symbol]
        return fq.Security(
            source=self.source,
            type="equity",
            **dataclasses.asdict(stock),
        )

    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[fq.Security, None, None]:
        """Fetch securities from the internet."""
        for symbol in symbols:
            try:
                yield self.fetch_security(symbol)
            except (KeyError, ValueError) as err:
                fq.log.warning("failed to fetch %s: %s", symbol, err)


def build_security_feed(*args: t.Any, **kwargs: t.Any) -> NasdaqSecurityFeed:
    """Build a security feed."""
    return NasdaqSecurityFeed(*args, **kwargs)



"""Retrieves data from Yahoo. Finance."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"

import contextlib
import datetime as dt
import decimal
import functools
import json
import re
import time
import typing as t
from decimal import Decimal as D  # noqa: N817

import requests
import requests.sessions

# Needed to be handle parsing times like 4:00pm and such.
import finquotes as fq

SOURCE = "finquotes.yahoo"

# Year threshold for 2-digit year conversion
YEAR_2000 = 2000

# See http://www.gummy-stuff.org/Yahoo-data.htm for more information.
fields_map = {
    fq.Field.SYMBOL: "symbol",
    fq.Field.NAME: "shortName",
    fq.Field.ASK: "ask",
    fq.Field.ASK_SIZE: "askSize",
    fq.Field.BID: "bid",
    fq.Field.BID_SIZE: "bidSize",
    fq.Field.TIME_STAMP: "regularMarketTime",
    fq.Field.DATE: "regularMarketTime",
    fq.Field.TIME: "regularMarketTime",
    fq.Field.EX_DATE: "dividenDate",
    fq.Field.VOLUME: "regularMarketVolume",
    fq.Field.EXCHANGE: "exchange",
    fq.Field.N_SHARES_OUTSTANDING: "sharesOutstanding",
    fq.Field.PCT_CHANGE: "regularMarketChangePercent",
    fq.Field.CHANGE: "regularMarketChange",
    fq.Field.PREVIOUS_CLOSE: "regularMarketPreviousClose",
    fq.Field.OPEN: "regularMarketOpen",
    fq.Field.HIGH: "regularMarketDayHigh",
    fq.Field.LOW: "regularMarketDayLow",
    fq.Field.CLOSE: "regularMarketPrice",
    fq.Field.CURRENCY: "currency",
}


def to_yahoo_timestamp(
    begin_date: dt.date,
    end_date: dt.date,
) -> tuple[str, str]:
    """Convert the date to as_of_date since epoch.

    Yahoo. stores data as timestamp of seconds sine epoch 1st Jan 1970.

    :param begin_ts: begin time stamp to convert to time stamp.
    :param end_ts: end time stamp to convert to time stamp.

    :return: time stamp of the day after date at 00:00 UTC.

    """
    begin_ts = dt.datetime.combine(begin_date, dt.time(), tzinfo=dt.UTC)
    end_ts = dt.datetime.combine(
        end_date,
        dt.time(),
        tzinfo=dt.UTC,
    ) + dt.timedelta(1)
    return f"{begin_ts.timestamp():.0f}", f"{end_ts.timestamp():.0f}"


def from_yahoo_timestamp(datestr: str) -> dt.datetime:
    """Convert a string representing an integer of seconds since epoch.

    Localize to yahoo.timezone.

    :return: datetime.datetime

    """
    try:
        n_secs = int(datestr)
        return dt.datetime.fromtimestamp(n_secs, tz=dt.UTC)
    except (ValueError, OverflowError) as err:
        fq.log.warning("failed to convert %s to a dt.datetime: %s", datestr, err)
        raise


# Convenience functions
def parse_date(datestr: str, **kw: t.Any) -> dt.date:
    """Parse a date retrieved from Yahoo."""
    matches = re.match(r"^(\d\d?)[/-]?(\d\d?)[/-]?(\d+)?$", datestr)
    date = None
    if matches:
        year = fq.today().year
        if matches.group(3):
            year = int(matches.group(3))
        if year < YEAR_2000:
            year += YEAR_2000
        month = matches.group(1)
        day = matches.group(2)
        with contextlib.suppress(ValueError):
            date = dt.date(int(year), int(month), int(day))
        if kw.get("as_of_date"):
            as_of_date = kw["as_of_date"]
            assert as_of_date <= fq.today()
            diff = fq.today() - as_of_date
            date = (date - diff).date()
    if not date:
        msg = f"cannot convert {datestr} to a date"
        raise TypeError(msg)
    return date


def replace_comma_within_prices(string: str) -> str:
    """Replace comma in the names of companies."""
    in_prices = False
    new_line = ""
    for char in string:
        if char == '"':
            in_prices = not in_prices
        new_char = " " if char == "," and in_prices else char
        new_line += new_char
    return new_line


def decode(raw_line: str) -> str:
    """Decode and strip spaces in argument.

    Assumes UTF-8 if the line is an instance of bytes.
    """
    line = raw_line.decode("utf-8") if isinstance(raw_line, bytes) else raw_line
    return str(line).strip()  # removes blanks.


class CrumbAndCookies(t.NamedTuple):
    """Container for crumb and cookies data."""

    crumb: str
    cookies: t.Any


@functools.lru_cache
def get_crumb_and_cookies(
    symbol: str,
    session: requests.sessions.Session,
) -> CrumbAndCookies:
    """Retrieve the main Yahoo. markers."""
    url = f"https://finance.yahoo.com/quote/{symbol}"
    content_text = fq.localhttp.request(session, "GET", url).text
    search = re.search(
        r'"CrumbStore":\{"crumb":"(.*)"\},"CompositeStore"',
        content_text,
    )
    if search is None:
        msg = "No crumb found in initial response"
        raise ValueError(msg)
    crumb = search.group(1)
    # For some reason does not get converted back to string.
    crumb = crumb.replace("\\u002F", "/")
    fq.log.info("using Yahoo. crumb %s", crumb)
    return CrumbAndCookies(crumb, session.cookies)


def make_session() -> requests.sessions.Session:
    """Get a context to retrieve internet data.

    Factors session customization needed to query Yahoo.

    """
    session = requests.sessions.Session()
    fq.localhttp.set_retries(session)
    return session


URLS = {
    "v7": {
        "quote": "https://query1.finance.yahoo.com/v7/finance/quote",
        "history": "https://query1.finance.yahoo.com/v7/finance/download/",
    },
}


class YahooFeedMixin(fq.NetworkConfigMixin):
    """Retrieves data from Yahoo."""

    markers = None
    size_re = re.compile(r"(,[1-9]\d?),(\d00)(\\r)")
    date_re = re.compile(r"((\d+)[-\s]+)?([A-Z][a-z]{2})[-\s]+(\d+)")

    def __init__(
        self: t.Self,
        session: requests.Session | None = None,
        crumb_and_cookies: CrumbAndCookies | None = None,
        *,
        adjust_highs: bool = False,
        adjust_lows: bool = False,
        version: str = "v7",
        max_retries: int = 3,
        retry_delay: float = 1.0,
        timeout: float = 30.0,
        **kwargs: t.Any,
    ) -> None:
        """Initialize YahooFeedMixin."""
        super().__init__(timeout=timeout, max_retries=max_retries, retry_delay=retry_delay, **kwargs)
        self.session = session or make_session()
        self.snap_url = URLS[version]["quote"]
        self.sec_url = URLS[version]["quote"]
        self.hist_url = URLS[version]["history"]
        self.fields = [
            fq.Field.SYMBOL,
            fq.Field.OPEN,
            fq.Field.HIGH,
            fq.Field.LOW,
            fq.Field.CLOSE,
            fq.Field.VOLUME,
            fq.Field.TIME_STAMP,
        ]
        self._crumb_and_cookies = crumb_and_cookies
        self.adjust_highs = adjust_highs
        self.adjust_lows = adjust_lows

    @property
    def crumb_and_cookies(self: t.Self) -> CrumbAndCookies:
        """Retrieve the main Yahoo. markers if not retrieved already."""
        if not self._crumb_and_cookies:
            self._crumb_and_cookies = get_crumb_and_cookies("SPY", self.session)
        return self._crumb_and_cookies

    @staticmethod
    def split_kwargs(kwargs: dict) -> tuple[dict, dict]:
        """Split kwargs between those known by YahooFeedMixin and everything else.

        Return:
        ------
            Tuple of yahoo keyword arguments, then the other keyword arguments.

        """
        yahoo_kwargs = {}
        yahoo_kwargs["crumb_and_cookies"] = kwargs.pop("crumb_and_cookies", None)
        yahoo_kwargs["session"] = kwargs.pop("session", make_session())
        for param in ["max_retries", "retry_delay", "timeout"]:
            if param in kwargs:
                yahoo_kwargs[param] = kwargs.pop(param)
        return yahoo_kwargs, kwargs

    def _request_text(self: t.Self, url: str, params: dict | None = None) -> str:
        """Forward call to _open_url with protection against HTTPError.

        Also check the status of the response. Logs the error as a warning and
        return an empty string as reponse if an issue occured.

        """
        text = ""
        for retry in range(self.max_retries):
            try:
                fq.log.debug(url)
                if retry > 0:
                    time.sleep(self.retry_delay)
                    fq.log.debug("trying %d of %d ...", retry, self.max_retries)
                text = fq.localhttp.request(
                    self.session,
                    "GET",
                    url,
                    params=params,
                ).text
                break
            except requests.exceptions.HTTPError as error:
                response = error.response
                if response is not None and response.status_code in {401, 404}:
                    msg = f"invalid url: {error}"
                    raise fq.FeedError(msg) from error
                fq.log.info("failed get for %s: %s", url, error)
                if retry >= self.max_retries - 1:
                    break
            except (
                ConnectionError,
                TimeoutError,
                requests.exceptions.Timeout,
                ValueError,
            ) as error:
                fq.log.warning("failed get for %s: %s", url, error)
                if retry >= self.max_retries - 1:
                    raise
        return text

    def get_cookies(self: t.Self) -> requests.cookies.RequestsCookieJar:
        """Get cookies for http call."""
        return self.crumb_and_cookies.cookies

    def _parse_line(self: t.Self, line: str) -> fq.Price:
        """Parse one line in the content returned by Yahoo."""
        line = line.strip()
        fq.log.debug(line)
        modline = replace_comma_within_prices(line)
        modline = modline.replace("&nbsp;", " ")
        values = modline.split(",")
        checked_values = {}
        for field, unstripped_value in zip(self.fields, values, strict=False):
            value = unstripped_value.strip('"').strip().rstrip("+").rstrip("-")
            try:
                checked_value: str | D | dt.date | dt.time | int | None = None
                if checked_value == "N/A":
                    pass
                elif field in [fq.Field.DATE, fq.Field.EX_DATE, fq.Field.PAY_DATE]:
                    checked_value = parse_date(value)
                elif field in [fq.Field.TIME, fq.Field.TIME_STAMP]:
                    checked_value = dt.datetime.strptime(value + "-04:00", "%I:%M%p%z").time()
                elif field not in [fq.Field.SYMBOL, fq.Field.NAME, fq.Field.EXCHANGE]:
                    checked_value = D(value)
            except (TypeError, ValueError, decimal.InvalidOperation) as err:
                fq.log.warning(
                    "failed to interpret %s as %s in %s %s",
                    value,
                    field,
                    line,
                    err,
                )
                checked_value = None
            if checked_value is not None:
                checked_values[field.name] = checked_value
        price = fq.Price(source=SOURCE, **checked_values)  # type: ignore[arg-type]
        fq.log.debug("%s", price)
        return price

    def _parse_csv(
        self: t.Self,
        content: list,
        symbols: list,
    ) -> t.Generator[fq.Price, None, None]:
        """Parse content returned by Yahoo."""
        for raw_line in content:
            line = decode(raw_line)
            # Silently skips empty lines.
            if not line:
                continue
            # Handles known errors.
            if line.startswith("Sorry"):
                fq.log.error("Server not available: %s", line)
                break
            try:
                data = self._parse_line(line)
            except ValueError as err:
                fq.log.error("Failed to parse %s: %s", line, str(err))
                continue
            if not data.symbol:
                fq.log.warning("no symbol found in %s", line)
                continue
            if data.symbol not in symbols:
                fq.log.warning("unexpected symbol %s in %s", data.symbol, line)
                continue
            if not data.date and "date" in self.fields:
                fq.log.warning("no date found in %s", line)
                continue
            yield data

    def _parse_json_price(
        self: t.Self,
        item: dict,
        fields: t.Iterable[fq.Field],
    ) -> fq.Price:
        """Parse single element in the JSON response looking for price info."""
        timestamp = from_yahoo_timestamp(item["regularMarketTime"])
        timestamp_date: dt.date | None = None
        # Set the date as a date() without time if time is not requested.
        if fq.Field.DATE in fields and fq.Field.TIME not in fields:
            timestamp_date = timestamp.date()
        assert timestamp is not None, "date is required either as DATE or TIME_STAMP"
        close = D(item["regularMarketPrice"])
        price = fq.Price(
            item["symbol"],
            date=timestamp_date or timestamp,
            close=close,
            source=SOURCE,
        )
        if fq.Field.OPEN in fields and "regularMarketOpen" in item and item["regularMarketOpen"] != "0.0":
            price.open = D(item["regularMarketOpen"])
        if fq.Field.HIGH in fields and "regularMarketDayHigh" in item and item["regularMarketDayHigh"] != "0.0":
            price.high = D(item["regularMarketDayHigh"])
        if fq.Field.LOW in fields and "regularMarketDayLow" in item and item["regularMarketDayLow"] != "0.0":
            price.low = D(item["regularMarketDayLow"])
        if fq.Field.VOLUME in fields and "regularMarketVolume" in item:
            price.volume = int(item["regularMarketVolume"]) if item["regularMarketVolume"] is not None else None
        return price

    def _parse_json_security(self: t.Self, content: dict) -> fq.Security:
        """Parse single elemenet in JSON response looking for security info."""
        name = content["shortName"]
        if "longName" in content:
            name = content["longName"]
        security = fq.Security(
            symbol=content["symbol"],
            name=name,
            source=SOURCE,
        )
        security.exchange = "X" + content["exchange"]
        security.type = content["quoteType"].lower()
        security.currency = content["currency"].upper()
        return security

    def _parse_results(self: t.Self, content: str) -> list:
        # No conversion of float because we handle the string later in Price
        # constructor with decimal.Decimal.
        fq.log.debug("%s", content)
        data = json.loads(content, parse_float=lambda x: x)
        if "quoteResponse" not in data:
            msg = f"no quoteRespose in {content}"
            raise ValueError(msg)
        data = data["quoteResponse"]
        if data.get("error"):
            fq.log.warning("error detected: %s", data["error"])
        if "result" not in data:
            msg = f"not result found in {data}"
            raise ValueError(msg)
        if not data["result"]:
            msg = f"result is empty in {data}"
            raise ValueError(msg)
        return data["result"]

    def _parse_json_prices(
        self: t.Self,
        content: str,
        _: t.Iterable[str],
        fields: t.Iterable[fq.Field],
    ) -> t.Generator[fq.Price, None, None]:
        """Parse json response looking for prices."""
        for result in self._parse_results(content):
            try:
                yield self._parse_json_price(result, fields)
            # OSError when regularMarketTime is 0.
            except (ValueError, KeyError, OSError) as err:
                fq.log.warning("failed to process %s: %s", json.dumps(result), err)

    def _parse_json_securities(
        self: t.Self,
        content: str,
    ) -> t.Generator[fq.Security, None, None]:
        """Parse json response looking for securities."""
        for result in self._parse_results(content):
            try:
                yield self._parse_json_security(result)
            except (ValueError, KeyError) as err:
                fq.log.warning("failed to process %s: %s", json.dumps(result), err)

    def _fetch_batch(
        self: t.Self,
        symbols: t.Iterable[str],
        fields: t.Iterable[fq.Field],
        mapped_fields: t.Iterable[str],
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve self.fields relative to symbols (internal use)."""
        symbolstr = ",".join(symbols)
        params = {
            "fields": mapped_fields,
            "symbols": symbolstr,
        }
        response = self._request_text(self.snap_url, params=params)
        if response:
            yield from self._parse_json_prices(response, symbols, fields)

    def fetch_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: None | dt.date = None,  # noqa: ARG002
        as_of_ts: None | dt.datetime = None,  # noqa: ARG002
        fields: t.Iterable[fq.Field] | None = None,
        batch_size: int = 195,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch information about the securities from Yahoo.

        :param symbols: list or single symbol because the same function is used
            for historical, and close prices.
        :param date: date for the prices looked up (not used, always as_of_date).
        :param as_of_ts: expected time stamp for the prices (not used).
        :param fields: list of fields being looked up (e.g. last, volume, ...)
        :param batch_size: max size of one request.

        Converts the argument to a list and call _fetch_batch to retrieve the
        data by block of 25 (Yahoo requirement).

        Note this function is both used as fetch_intraday_prices and
        fetch_close_prices.

        """
        if fields is None:
            fields = self.fields
        dedup_symbols = list(dict.fromkeys(symbols))
        mapped_fields = ",".join([fields_map[fld] for fld in fields])
        for idx in range(0, len(dedup_symbols), batch_size):
            yield from self._fetch_batch(
                dedup_symbols[idx : idx + batch_size],
                fields,
                mapped_fields,
            )

    def fetch_hist_dists(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Distribution, None, None]:
        """Fetch distribution history from Yahoo.

        If begin_date is not specified, the algorithm uses end_date.
        If end_date is not specified, the algorithm uses today.

        Returns the number of distributions succesfully retrieved.
        """
        if end_date is None:
            end_date = fq.today()
        if begin_date is None:
            begin_date = end_date
        begin, end = to_yahoo_timestamp(begin_date, end_date)
        params = {"period1": begin, "period2": end, "interval": "1d", "events": "div"}
        url = self.hist_url + symbol
        response = self._request_text(url, params=params)
        for raw_line in response.split("\n"):
            line = decode(raw_line).strip()
            if not raw_line:
                continue
            fq.log.debug("retrieved data: %s", line)
            vals = line.split(",")
            if vals[0] == "Date":
                continue  # Header line.
            try:
                date = dt.date.fromisoformat(vals[0])
            except ValueError as err:
                msg = "{}: cannot convert {} to a date: {}"
                msg = msg.format(line, vals[0], err)
                fq.log.warning(msg)
                continue
            if date > end_date or date < begin_date:
                msg = "{}: distribution date not between {} and {}"
                msg = msg.format(line, begin_date, end_date)
                fq.log.warning(msg)
                continue
            yield fq.Distribution(symbol, date, D(vals[1]), None)

    def fetch_hist_splits(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Split, None, None]:
        """Fetch distribution history from Yahoo.

        If begin_date is not specified, the algorithm uses end_date.
        If end_date is not specified, the algorithm uses today.

        Returns the number of distributions succesfully retrieved.
        """
        if end_date is None:
            end_date = fq.today()
        if begin_date is None:
            begin_date = end_date
        begin, end = to_yahoo_timestamp(begin_date, end_date)
        params = {"period1": begin, "period2": end, "interval": "1d", "events": "split"}
        url = self.hist_url + symbol
        response = self._request_text(url, params=params)
        for raw_line in response.split("\n"):
            line = decode(raw_line)
            # Silently skips empty lines.
            if not line:
                continue
            fq.log.debug("retrieved data: %s", line)
            vals = line.split(",")
            if vals[0] == "Date":
                continue  # Header line.
            try:
                date = dt.date.fromisoformat(vals[0])
            except ValueError as err:
                msg = "{}: cannot convert {} to a date: {}"
                msg = msg.format(line, vals[0], err)
                fq.log.warning(msg)
                continue
            if date > end_date or date < begin_date:
                msg = "{}: distribution date not between {} and {}"
                msg = msg.format(line, begin_date, end_date)
                fq.log.warning(msg)
                continue
            match = re.match(r"(\d+)\D(\d+)", vals[1])
            if match is None:
                msg = f"cannot interpret {vals[1]} as split"
                raise ValueError(msg)
            newq, oldq = match.group(1), match.group(2)
            yield fq.Split(symbol, date, int(newq), int(oldq))

    def fetch_securities(
        self: t.Self,
        symbols: t.Iterable[str],
    ) -> t.Generator[fq.Security, None, None]:
        """Fetch a security from Yahoo."""
        if symbols:
            symbolstr = ",".join(symbols)
            params = {
                "symbols": symbolstr,
            }
            response = self._request_text(self.snap_url, params=params)
            if not response:
                fq.log.error("no data returned for %s", symbolstr)
                msg = f"no data returned for {symbolstr}"
                raise fq.FeedError(msg)
            yield from self._parse_json_securities(response)


class YahooCloseFeed(YahooFeedMixin, fq.AbstractCloseFeed):
    """Implements finquotes.CloseFeed with YahooFeedMixin."""

    def __init__(self: t.Self, /, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
        as_of_ts: dt.datetime | None = None,
        fields: t.Iterable[fq.Field] | None = None,
        batch_size: int = 195,
        **kwargs: t.Any,
    ) -> t.Generator[fq.Price, None, None]:
        """Adapts fetch_prices to fetch_close_prices function."""
        return YahooFeedMixin.fetch_prices(
            self,
            symbols,
            date,
            as_of_ts,
            fields,
            batch_size,
            **kwargs,
        )


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> YahooCloseFeed:
    """Forward to build_feed."""
    return YahooCloseFeed(*args, **kwargs)


class YahooHistFeed(YahooFeedMixin, fq.AbstractHistoricalFeed):
    """Implements finquotes.AbstractHistFeed via Yahoo."""

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def _validate_date_range(
        self,
        as_of_date: dt.date,
        begin_date: dt.date,
        end_date: dt.date,
    ) -> bool:
        """Validate if the date is within the specified range."""
        if as_of_date > end_date:
            fq.log.warning(
                "ignored price for %s later than end date %s",
                as_of_date,
                end_date,
            )
            return False
        if as_of_date < begin_date:
            fq.log.warning(
                "ignored price for %s earlier than begin date %s",
                as_of_date,
                begin_date,
            )
            return False
        return True

    def _calculate_split_factor(
        self,
        splits: t.Sequence[fq.Split] | None,
        as_of_date: dt.date,
    ) -> D:
        """Calculate the split factor for the given date."""
        split_factor = D("1.0")
        if splits is not None:
            for split in splits:
                assert isinstance(split, fq.Split)
                if split.ex_date > as_of_date:
                    split_factor *= D(split.new_quantity / split.old_quantity)
        return split_factor

    def _create_price_object(
        self,
        symbol: str,
        as_of_date: dt.date,
        vals: list[str],
        split_factor: D,
    ) -> fq.Price:
        """Create and populate a Price object with the given data."""
        close = D.quantize(D(vals[4]) * split_factor, D("0.0000"))
        price = fq.Price(symbol, as_of_date, close, source=SOURCE)
        # Open, high, and low are dividend adjusted but not split adjusted.
        # http://blog.fosstrading.com/2017/06/yahoo-finance-alternatives.html
        price.open = D.quantize(D(vals[1]) * split_factor, D("0.0000"))
        price.high = D.quantize(D(vals[2]) * split_factor, D("0.0000"))
        price.low = D.quantize(D(vals[3]) * split_factor, D("0.0000"))
        price.volume = int(int(vals[6]) / float(split_factor)) if vals[6] is not None else None
        return price

    def _adjust_price_bounds(self, price: fq.Price) -> None:
        """Adjust price high and low values if needed."""
        if self.adjust_lows and price.low is not None and price.low > price.close:
            fq.log.warning(
                "adjusting low %s to close price %s",
                price.low,
                price.close,
            )
            price.low = price.close
        if self.adjust_highs and price.high is not None and price.high < price.close:
            fq.log.warning(
                "adjusting high %s to close price %s",
                price.high,
                price.close,
            )
            price.high = price.close

    def fetch_hist_price(
        self: t.Self,
        raw_line: str,
        symbol: str,
        begin_date: dt.date,
        end_date: dt.date,
        distributions: t.Sequence[fq.Distribution] | None = None,
        splits: t.Sequence[fq.Split] | None = None,
    ) -> fq.Price | None:
        """Convert string to price."""
        assert distributions is None or hasattr(distributions, "__iter__")
        assert splits is None or hasattr(splits, "__iter__")
        line = decode(raw_line)
        fq.log.debug("retrieved data: %s", line)
        # Silently skips empty lines.
        if not line:
            return None
        vals = line.split(",")
        as_of_date = dt.date.fromisoformat(vals[0])

        if not self._validate_date_range(as_of_date, begin_date, end_date):
            return None

        split_factor = self._calculate_split_factor(splits, as_of_date)
        price = self._create_price_object(symbol, as_of_date, vals, split_factor)
        self._adjust_price_bounds(price)
        return price

    # pylint: disable=R0913
    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        splits: t.Sequence[fq.Split] | None = None,
        distributions: t.Sequence[fq.Distribution] | None = None,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch history from Yahoo.

        :param begin_date: if not specified, the algorithm uses end_date.
        :param end_date: if not specified, the algorithm uses today.
        :param frequency: defaults to d[aily].  Can be w[eekly] or m[onthly].

        :return: iterable on finquotes.Price

        """
        assert symbol is not None, "symbol argument is None"
        assert isinstance(
            symbol,
            str,
        ), f"got symbol of type {type(symbol)} expected str"
        end_date, begin_date = fq.compute_default_dates(end_date, begin_date)
        if not frequency:
            frequency = "d"
        if frequency[0] not in ["d", "w", "m"]:
            msg = f"invalid frequency {frequency}"
            raise TypeError(msg)
        begin, end = to_yahoo_timestamp(begin_date, end_date)
        params = {
            "period1": begin,
            "period2": end,
            "interval": "1" + frequency[0],
            "events": "history",
        }
        url = self.hist_url + symbol
        response = self._request_text(url, params=params)
        for raw_line in response.split("\n"):
            if raw_line.startswith("Date"):
                continue
            if "null" in raw_line:
                continue
            if not raw_line:
                continue
            try:
                price = self.fetch_hist_price(
                    raw_line,
                    symbol,
                    begin_date,
                    end_date,
                    splits=splits,
                    distributions=distributions,
                )
            except ValueError as err:
                fq.log.warning(err)
                continue
            if price is None:
                continue
            yield price


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> YahooHistFeed:
    """Forward to build_feed."""
    return YahooHistFeed(*args, **kwargs)


class YahooDistFeed(YahooFeedMixin, fq.AbstractDistributionFeed):
    """Implements finquotes.AbstractDistFeed with YahooFeedMixin."""

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    fetch_hist_dists = YahooFeedMixin.fetch_hist_dists


def build_distribution_feed(*args: t.Any, **kwargs: t.Any) -> YahooDistFeed:
    """Forward to build_feed."""
    return YahooDistFeed(*args, **kwargs)


class YahooSplitFeed(YahooFeedMixin, fq.AbstractSplitFeed):
    """Implements finquotes.AbstractSplitFeed with YahooFeedMixin."""

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    fetch_hist_splits = YahooFeedMixin.fetch_hist_splits


def build_split_feed(*args: t.Any, **kwargs: t.Any) -> YahooSplitFeed:
    """Forward to build_feed."""
    return YahooSplitFeed(*args, **kwargs)


class YahooSecurityFeed(YahooFeedMixin, fq.AbstractSecurityFeed):
    """Implements finquotes.AbstractSecurityFeed with YahooFeedMixin."""

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    fetch_securities = YahooFeedMixin.fetch_securities


def build_security_feed(*args: t.Any, **kwargs: t.Any) -> YahooSecurityFeed:
    """Forward to build_feed."""
    return YahooSecurityFeed(*args, **kwargs)



"""Command line interface useful for testing."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"

import datetime as dt
import logging.config
import sys
import typing as t

import click

import finquotes as fq

LOGGING_CONFIG: dict[str, t.Any] = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "warning": {"format": "%(asctime)s [%(levelname)s]: %(message)s"},
        "info": {"format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"},
        "debug": {
            "format": "%(asctime)s [%(levelname)s] %(name)s %(pathname)s:%(lineno)d: %(message)s",
        },
    },
    "handlers": {
        "console": {
            "formatter": "warning",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",  # Default is stderr
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["console"],
            "level": "WARNING",
        },
        "finquotes": {"propagate": True},
        "requests": {"propagate": True},
        "urllib3": {"propagate": True},
        "__main__": {
            "propagate": False,
        },
    },
}

log = logging.getLogger("finquotes")


def str_to_feed_type_enum(feed: str) -> fq.FeedType:
    """Convert str to finquotes.FeedType."""
    try:
        return fq.FeedType[feed.upper()]
    except KeyError:
        types = " ".join(str(f).split(".")[1].lower() for f in fq.FeedType)
        fq.log.error("%s is not a valid feed type. Pick from %s", feed, types)
        raise


class _SymbolsNotSeen:
    """Helper class to mark a symbol as not seen."""

    def __init__(self: t.Self, symbols: list[str]) -> None:
        self.symbols = set(symbols)

    def __str__(self: t.Self) -> str:
        return ", ".join(self.symbols)

    def to_str(
        self: t.Self,
        price: fq.Price | fq.Security | fq.Split | fq.Distribution,
        *,
        discard: bool = True,
    ) -> str:
        """Mark a symbol as seen and return its string representation."""
        if discard:
            self.symbols.discard(price.symbol)
        return str(price)


def _fetch_close_data(
    feed: "fq.AbstractCloseFeed",
    symbols: list[str],
    end_date: dt.date,
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch close price data."""
    return [not_seen.to_str(price) for price in feed.fetch_close_prices(symbols, date=end_date)]


def _fetch_security_data(
    feed: "fq.AbstractSecurityFeed",
    symbols: list[str],
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch security data."""
    return [not_seen.to_str(security) for security in feed.fetch_securities(symbols)]


def _fetch_historical_data(
    feed: "fq.AbstractHistoricalFeed",
    symbols: list[str],
    begin_date: dt.date,
    end_date: dt.date,
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch historical price data."""
    output: list[str] = []
    for sym in symbols:
        output.extend(
            not_seen.to_str(price)
            for price in feed.fetch_hist_prices(
                sym,
                begin_date=begin_date,
                end_date=end_date,
            )
        )
    return output


def _fetch_split_data(
    feed: "fq.AbstractSplitFeed",
    symbols: list[str],
    begin_date: dt.date,
    end_date: dt.date,
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch split data."""
    output: list[str] = []
    for sym in symbols:
        output.extend(
            not_seen.to_str(price)
            for price in feed.fetch_hist_splits(
                sym,
                begin_date=begin_date,
                end_date=end_date,
            )
        )
    return output


def _fetch_distribution_data(
    feed: "fq.AbstractDistributionFeed",
    symbols: list[str],
    begin_date: dt.date,
    end_date: dt.date,
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch distribution data."""
    output: list[str] = []
    for sym in symbols:
        output.extend(
            not_seen.to_str(price)
            for price in feed.fetch_hist_dists(
                sym,
                begin_date=begin_date,
                end_date=end_date,
            )
        )
    return output


def _fetch_data_by_feed_type(
    feed_type: fq.FeedType,
    feed: t.Any,
    symbols: list[str],
    begin_date: dt.date,
    end_date: dt.date,
    not_seen: "_SymbolsNotSeen",
) -> list[str]:
    """Fetch data based on feed type."""
    if feed_type == fq.FeedType.CLOSE:
        close_feed = t.cast("fq.AbstractCloseFeed", feed)
        return _fetch_close_data(close_feed, symbols, end_date, not_seen)
    if feed_type == fq.FeedType.SECURITY:
        security_feed = t.cast("fq.AbstractSecurityFeed", feed)
        return _fetch_security_data(security_feed, symbols, not_seen)
    if feed_type == fq.FeedType.HISTORICAL:
        historical_feed = t.cast("fq.AbstractHistoricalFeed", feed)
        return _fetch_historical_data(historical_feed, symbols, begin_date, end_date, not_seen)
    if feed_type == fq.FeedType.SPLIT:
        split_feed = t.cast("fq.AbstractSplitFeed", feed)
        return _fetch_split_data(split_feed, symbols, begin_date, end_date, not_seen)
    if feed_type == fq.FeedType.DISTRIBUTION:
        distribution_feed = t.cast("fq.AbstractDistributionFeed", feed)
        return _fetch_distribution_data(distribution_feed, symbols, begin_date, end_date, not_seen)
    log.error("feed type %s not handled right now", feed_type)
    return []


def protected_cli(
    feed_name: str,
    symbols_iter: t.Iterable[str],
    source: str,
    end_datetime: dt.datetime | None,
    begin_datetime: dt.datetime | None,
    **kwargs: t.Any,
) -> None:
    """Inner function may throw."""
    feed_type = str_to_feed_type_enum(feed_name)
    feed = fq.build_feed(source, feed_type=feed_type, **kwargs)
    end_date, begin_date = fq.compute_default_dates(
        end_datetime.date() if end_datetime else None,
        begin_datetime.date() if begin_datetime else None,
    )
    symbols = list(symbols_iter)
    if len(symbols) == 0:
        msg = "at least one symbol required.  See --help"
        raise click.UsageError(msg)

    not_seen = _SymbolsNotSeen(symbols)
    output = _fetch_data_by_feed_type(feed_type, feed, symbols, begin_date, end_date, not_seen)

    for line in sorted(output):
        click.echo(line)
    if not_seen.symbols:
        msg = f"symbol(s) not found for feed {feed_name} between date {begin_date} and {end_date}: {not_seen}"
        raise ValueError(msg)


def setup_logging(*, verbose: bool) -> None:
    """Set up logging based on verbose level.

    :param: verbose 0 (WARNING, default), 1 (INFO) or 2 (DEBUG) is the number of -v.

    """
    logging.config.dictConfig(LOGGING_CONFIG)
    level = logging.WARNING - max(0, min(2, verbose)) * 10
    log.setLevel(level)
    level_name = logging.getLevelName(level).lower()
    logging_format = LOGGING_CONFIG["formatters"][level_name]["format"]
    logging.getLogger().handlers[0].setFormatter(logging.Formatter(logging_format))


@click.command(name="python -m finquotes")
@click.argument(
    "feed",
    # rename close to snap or snapshot everywhere?
    type=click.Choice(["close", "historical", "distribution", "split", "security"]),
)
@click.argument(
    "symbols",
    nargs=-1,
    metavar="SYMBOLS",
)
@click.option(
    "-s",
    "--source",
    help="price source to use",
    metavar="MODULE",
    default="finquotes.yahoo",
)
@click.option(
    "-d",
    "--end-date",
    " --date",
    help="end of period of interest",
    type=click.DateTime(["%Y-%m-%d"]),
)
@click.option(
    "-b",
    "--begin-date",
    help="start of period of interest",
    type=click.DateTime(["%Y-%m-%d"]),
)
@click.option(
    "-v",
    "--verbose",
    help="increases verbosity",
    count=True,
)
@click.option("--api-key", help="set API key for the specific source")
@click.option("--debug", is_flag=True, help="raise exceptions on error")
@click.option("--timeout", type=float, help="set timeout in seconds")
@click.option("--max-retries", type=int, help="set max retries")
@click.option("--retry-delay", type=float, help="set retry delay in seconds")
@click.version_option(version=fq.__version__, message="v%(version)s")
def main(
    feed: str,
    symbols: t.Iterable[str],
    source: str,
    end_date: dt.datetime,
    begin_date: dt.datetime,
    verbose: int,
    **kwargs: t.Any,
) -> None:
    """Command line interface for finquotes is useful for testing."""
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    debug = kwargs.pop("debug")
    try:
        setup_logging(verbose=verbose > 0)
        protected_cli(
            feed,
            symbols,
            source,
            end_date,
            begin_date,
            **kwargs,
        )
    except Exception as error:  # pylint: disable=broad-except
        if debug:
            raise
        # Otherwise, only show the error.
        log.info(
            "error detected and absorbed, use --debug option to show the stack trace",
        )
        log.exception("%s", type(error).__name__)
        click.get_current_context().exit(1)


if __name__ == "__main__":
    main(sys.argv[1:])  # pylint: disable=no-value-for-parameter



"""Retrieve data from http://www.barchartondemand.com."""

__author__ = "Jerome Lecomte"
__license__ = "MIT"


import datetime as dt
import functools
import json
import typing as t
from collections.abc import Sequence
from decimal import Decimal as D  # noqa: N817

import requests
from dateutil import parser as dtparser

import finquotes as fq

SOURCE = "finquotes.barchart"
get_api_key = functools.partial(fq.get_api_key, api_name=SOURCE)


class BarChartMixin(fq.NetworkConfigMixin):
    """Get historical price from barchart."""

    base_url = "http://marketdata.websol.barchart.com"
    index_url = "https://core-api.barchart.com/v1/quotes/get"

    def __init__(
        self: t.Self,
        api_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(**kwargs)
        self.batch_size = 99
        self.api_key = api_key
        if self.api_key is None:
            self.api_key = get_api_key()

    def get_name(self: t.Self) -> str:
        """Return barchart as the symbolic name."""
        return SOURCE

    def parse_hist_prices_line(
        self: t.Self,
        symbol: str,
        line: str,
    ) -> t.Generator[fq.Price, None, None]:
        """Parse line returned by barchart API."""
        vals = line.split(",")
        vals = [val.strip()[1:-1] for val in vals]  # remove quotes.
        date = dt.date.fromisoformat(vals[2])
        if symbol != vals[0]:
            msg = f"found value {vals[0]} where {symbol} expected in {line}"
            raise ValueError(msg)
        if not vals[6]:
            msg = f"price not found in {line}"
            raise ValueError(msg)
        price = fq.Price(
            symbol=symbol,
            date=date,
            open_=D(vals[3]),
            high=D(vals[4]),
            low=D(vals[5]),
            close=D(vals[6]),
            volume=fq.to_int_or_none(vals[7]),
            source=self.get_name(),
        )
        yield price

    def parse_close_prices_line(
        self: t.Self,
        line: str,
    ) -> t.Generator[fq.Price, None, None]:
        """Parse line returned by barchart API."""
        vals = line.split(",")
        vals = [val.strip()[1:-1] for val in vals]  # remove quotes.
        vals[7] = vals[7][:-6]  # remove time zone for date.
        date = dt.datetime.strptime(vals[7] + "+00:00", "%Y-%m-%dT%H:%M:%S%z").date()
        price = fq.Price(
            symbol=vals[0],
            date=date,
            open_=D(vals[11]),
            high=D(vals[12]),
            low=D(vals[13]),
            close=D(vals[14]),
            volume=fq.to_int_or_none(vals[16]),
            source=self.get_name(),
        )
        yield price

    def fetch_close_prices_batched(
        self: t.Self,
        symbols: list[str],
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve current prices for a batch.

        Note that client should call `fetch_close_prices` instead of this
        function directly.

        :param list(str) symbols: list of `self.batch_size` symbols or less.

        """
        assert len(symbols) <= self.batch_size, "too many symbols"
        url = f"{self.base_url}/getQuote.csv"
        params = {"key": self.api_key or "", "symbols": ",".join(symbols)}
        params = {k: v for k, v in params.items() if v is not None}
        params = {k: str(v) for k, v in params.items()}
        with requests.Session() as session:
            response_text = fq.localhttp.get(session, url, params=params)
        for line in response_text.split("\n"):
            if not line or line.startswith("symbol"):
                continue
            try:
                yield from self.parse_close_prices_line(line)
            except (IndexError, TypeError, ValueError) as err:
                fq.log.warn("failed to process %d: %s", line, err)

    def __fetch_index_prices_in(
        self: t.Self,
        page: str,
        symbols: set[str],
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve prices in a partcular barchart page.

        :param str page: page name to fetch.
        :param set(str) symbols: remaining symbols to be fetched.

        """
        params = {
            "list": page,
            "fields": (
                "symbol,symbolName,lastPrice,priceChange,openPrice,highPrice,lowPrice,tradeTime,symbolCode,symbolType"
            ),
            "orderBy": "",
            "orderDir": "desc",
            "page": "1",
            "limit": "100",
            "raw": "0",
        }
        params = {k: v for k, v in params.items() if v is not None}
        params = {k: str(v) for k, v in params.items()}
        with requests.Session() as session:
            response_text = fq.localhttp.get(session, self.index_url, params)
        content = json.loads(response_text)
        if content["errors"]:
            fq.log.warning(content["errors"])
        data = content["data"]
        for value in data:
            if value["symbol"] not in symbols:
                continue
            try:
                trade_time = dtparser.parse(value["tradeTime"])
                if not trade_time.tzinfo:
                    trade_time = trade_time.replace(tzinfo=dt.UTC)
            except (ValueError, TypeError, OverflowError) as error:
                fq.log.info(
                    "failed to parse %s as datetime: %s",
                    value["tradeTime"],
                    error,
                )
                continue
            symbol = value["symbol"]
            price = fq.Price(
                symbol=symbol,
                date=trade_time.date(),
                close=value["lastPrice"].replace(",", ""),
                open_=value["openPrice"].replace(",", ""),
                low=value["lowPrice"].replace(",", ""),
                high=value["highPrice"].replace(",", ""),
                source=SOURCE,
            )
            symbols.remove(symbol)
            yield price

    def fetch_index_prices(
        self: t.Self,
        symbols: set[str],
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve index prices.

        :param set(str) symbols: list of symols (all indexes)

        """
        pages = ["stocks.indices.us", "stocks.markets.volatility"]
        for page in pages:
            if len(symbols) == 0:
                break
            yield from self.__fetch_index_prices_in(page, symbols)

    @staticmethod
    def is_index(symbol: str) -> bool:
        """Check if the symbol is an index."""
        return symbol.startswith("$")


class BarChartHistFeed(BarChartMixin, fq.AbstractHistoricalFeed):
    """Retrieve historical prices from barchart."""

    def __init__(self: t.Self, api_key: str | None = None, **kwargs: t.Any) -> None:
        super().__init__(api_key=api_key, name=__name__, source=SOURCE, **kwargs)

    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,  # noqa: ARG002
        splits: Sequence[fq.Split] | None = None,  # noqa: ARG002
        distributions: Sequence[fq.Distribution] | None = None,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve historical prices.

        :param str symbol: symbol of the price serie.
        :param end_date: end of the period.
        :param begin_date: start of the period.
        :param splits: ignored.
        :param distributions: ignored
        :return: list(finquotes.HistoricalPrice)

        """
        end_date, begin_date = fq.compute_default_dates(end_date, begin_date)
        url = f"{self.base_url}/getHistory.csv"
        params = {
            "key": self.api_key or "",
            "symbol": symbol,
            "type": "daily",
            "startDate": begin_date.strftime("%Y%m%d%H%M%S"),
            "endDate": end_date.strftime("%Y%m%d%H%M%S"),
            "splits": "false",
            "dividends": "false",
        }
        params = {k: v for k, v in params.items() if v is not None}
        params = {k: str(v) for k, v in params.items()}
        with requests.Session() as session:
            response_text = fq.localhttp.get(session, url, params=params)
        for line in response_text.split("\n"):
            if not line or line.startswith("symbol"):
                continue
            if "maximum number of requests" in line:
                raise RuntimeError(line)
            try:
                yield from self.parse_hist_prices_line(symbol, line)
            except (IndexError, TypeError, ValueError) as err:
                fq.log.warning("failed to process %d: %s", line, err)


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> BarChartHistFeed:
    """Return a BarChartHistFeed instance."""
    return BarChartHistFeed(*args, **kwargs)


class BarChartCloseFeed(BarChartMixin, fq.AbstractCloseFeed):
    """Retrieve current prices from barchart."""

    def __init__(
        self: t.Self,
        api_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        super().__init__(api_key=api_key, name=__name__, source=SOURCE, **kwargs)

    def fetch_close_prices(
        self: t.Self,
        symbols: Sequence[str],
        date: dt.date | None = None,  # noqa: ARG002
        **kwargs: t.Any,
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve current prices.

        :param symbols: symbols looked up.

        Yield a generator iterating over `finquotes.Price`.

        """

        def is_index(symbol: str) -> bool:
            """Check if the symbol is an index."""
            return symbol.startswith("$")

        index_syms, stock_syms = set(), []
        for symbol in symbols:
            if is_index(symbol):
                index_syms.add(symbol)
            else:
                stock_syms.append(symbol)
        yield from self.fetch_index_prices(index_syms, **kwargs)
        for start in range(0, len(stock_syms), self.batch_size):
            finish = start + self.batch_size
            yield from self.fetch_close_prices_batched(
                stock_syms[start:finish],
                **kwargs,
            )


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> BarChartCloseFeed:
    """Return a BarChartCloseFeed instance."""
    return BarChartCloseFeed(*args, **kwargs)



"""Interface to stooq.pl.

Provides end-of-day data for stocks and indexes (no ETF)

:seealso: https://stooq.com/db/h/

"""

__author__ = "Jerome Lecomte"
__license__ = "MIT"


import datetime as dt
import typing as t
from decimal import Decimal as D, InvalidOperation  # noqa: N817

import bs4  # type: ignore[import-untyped]
import requests

import finquotes as fq

SOURCE = "finquotes.stooq"


def _open_url(*args: t.Any, **kwargs: t.Any) -> str:
    """Forward the call to requests, check status and decode text."""
    # Extract timeout to make it explicit for mypy
    timeout = kwargs.pop("timeout", 30.0)
    response = requests.get(*args, timeout=timeout, **kwargs)
    fq.log.debug("fetched %s", response.url)
    response.raise_for_status()
    if response.encoding is None:
        response.encoding = "utf-8"
    return response.content.decode(response.encoding)


def parse_polish_date(pol_date_str: str) -> dt.date:
    """Convert a polish date to datetime.datetime.

    We don't rely on locale because it seems that the local needs to be
    installed on the host and we just need to translate week days and months.

    """
    month_translations = {
        "cze": 6,
        "lis": 11,
        "lut": 2,
        "maj": 5,
        "sie": 8,
    }
    _, day, month, year = pol_date_str.split()
    return dt.date(int(year), month_translations[month], int(day))


class StooqFeed(fq.NetworkConfigMixin, fq.AbstractHistoricalFeed, fq.AbstractDistributionFeed):
    """Historical close prices from stooq.pl."""

    base_url = "https://stooq.com/q/"

    def __init__(self: t.Self, **kwargs: t.Any) -> None:
        super().__init__(name=__name__, source=SOURCE, **kwargs)

    def fetch_hist_price(
        self: t.Self,
        symbol: str,
        line: str,
    ) -> t.Generator[fq.Price, None, None]:
        """Parse specific line and return finquotes.Price."""
        datestr, open_, high, low, close, *others = line.split(",")
        date = dt.date.fromisoformat(datestr)
        price = fq.Price(
            source=self.source,
            symbol=symbol,
            date=date,
            close=close,
        )
        price.open = fq.to_decimal_or_none(open_)
        price.high = fq.to_decimal_or_none(high)
        price.low = fq.to_decimal_or_none(low)
        if others:
            price.volume = int(others[0]) if others[0] else None
        yield price

    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        splits: t.Sequence[fq.Split] | None = None,  # noqa: ARG002
        distributions: t.Sequence[fq.Distribution] | None = None,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Retrieve price history.

        :param symbol: ticker of the security.
        :param begin_date: if not specified, the algorithm uses end_date.
        :param end_date: if not specified, the algorithm uses today.
        :param frequency: defaults to d[aily].  Can be w[eekly] or m[onthly].

        :return: generator of Price objects

        """
        end_date, begin_date = fq.compute_default_dates(end_date, begin_date)
        if frequency is None:
            frequency = "d"
        date_format = "%Y%m%d"
        params = {
            "s": symbol,
            "i": frequency,
            "d1": begin_date.strftime(date_format),
            "d2": end_date.strftime(date_format),
        }
        url = self.base_url + "d/l/"
        content = _open_url(url, params=params, timeout=self.timeout)
        for line in content.split("\n"):
            if not line or line.startswith("Date"):
                continue
            try:
                yield from self.fetch_hist_price(symbol, line)
            except ValueError:
                fq.log.warning("%s: failed to process line: %s", symbol, line)

    def fetch_hist_dists(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
    ) -> t.Generator[fq.Distribution, None, None]:
        """Retrieve distribution history (dividend, capital gains, ...).

        :param symbol: ticker of the security.
        :param end_date: if not specified, the algorithm uses today.
        :param begin_date: if not specified, the algorithm uses end_date.

        :return: generator of Distribution objects

        NOTE: The distribution returned by stooq are adjusted and should
        not be used as actual as-of distributions.
        See https://stooq.com/pomoc/?q=9 (w/ google translate)

        """
        end_date, begin_date = fq.compute_default_dates(end_date, begin_date)
        params = {
            "s": symbol,
        }
        url = self.base_url + "m/"
        content = _open_url(url, params=params, timeout=self.timeout)
        soup = bs4.BeautifulSoup(content, "html.parser")
        table = soup.find("table", attrs={"id": "fth1"})
        if table is None:
            fq.log.warning("No table with id 'fth1' found for symbol: %s", symbol)
            return
        tbody = table.find("tbody")  # type: ignore[attr-defined]
        if tbody is None:
            fq.log.warning("No tbody found in table for symbol: %s", symbol)
            return
        rows = tbody.find_all("tr")  # type: ignore[attr-defined]
        for row in rows:
            cols = [el.text for el in row.find_all("td")]  # type: ignore[attr-defined]
            ex_date, amount = None, None
            try:
                ex_date = parse_polish_date(cols[0])
            except (ValueError, TypeError, IndexError) as err:
                fq.log.warning("failed to convet %s to date: %s", cols[0], err)
            try:
                amount = D(cols[2])
            except (ValueError, TypeError, IndexError, InvalidOperation) as err:
                fq.log.warning("failed to convert %s to decimal: %s", cols[2], err)
            if ex_date and amount:
                dist = fq.Distribution(symbol=symbol, ex_date=ex_date, amount=amount)
                yield dist


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> StooqFeed:
    """Forward to build_feed."""
    return StooqFeed(*args, **kwargs)



"""Network configuration mixin for feed classes."""

import typing as t


class NetworkConfigMixin:
    """Mixin that provides default implementations of network configuration properties.

    This class provides concrete implementations of the abstract network properties
    defined in AbstractFeed. Feeds can inherit from this mixin to get default
    implementations rather than implementing these properties individually.
    """

    NETWORK_CONFIG_KWARGS: t.ClassVar[list[str]] = ["timeout", "max_retries", "retry_delay", "api_key"]

    @classmethod
    def separate_network_kwargs(cls, kwargs: dict[str, t.Any]) -> tuple[dict[str, t.Any], dict[str, t.Any]]:
        """Separate network configuration parameters from other parameters.

        :param kwargs: Dictionary of keyword arguments

        :return: Tuple of (network_kwargs, other_kwargs)

        """
        network_kwargs = {}
        other_kwargs = {}
        for key, value in kwargs.items():
            if key in cls.NETWORK_CONFIG_KWARGS:
                network_kwargs[key] = value
            else:
                other_kwargs[key] = value

        return network_kwargs, other_kwargs

    def __init__(
        self: t.Self,
        timeout: float = 30.0,
        max_retries: int = 3,
        retry_delay: float = 1.0,
        api_key: str | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize network configuration.

        :param timeout: Network timeout in seconds
        :param max_retries: Maximum number of retries for failed requests
        :param retry_delay: Delay between retries in seconds
        :param api_key: API key for authentication
        :param kwargs: Additional arguments passed to parent classes
        """
        # Initialize our own attributes
        self._timeout = timeout
        self._max_retries = max_retries
        self._retry_delay = retry_delay
        self._api_key = api_key

        # Call super() to pass remaining kwargs to next class in MRO
        super().__init__(**kwargs)

    @property
    def timeout(self: t.Self) -> float:
        """Get timeout in seconds for network operations."""
        return self._timeout

    @timeout.setter
    def timeout(self: t.Self, value: float) -> None:
        """Set timeout in seconds for network operations."""
        self._timeout = value

    @property
    def max_retries(self: t.Self) -> int:
        """Get maximum number of retries for network operations."""
        return self._max_retries

    @max_retries.setter
    def max_retries(self: t.Self, value: int) -> None:
        """Set maximum number of retries for network operations."""
        self._max_retries = value

    @property
    def retry_delay(self: t.Self) -> float:
        """Get delay between retries in seconds."""
        return self._retry_delay

    @retry_delay.setter
    def retry_delay(self: t.Self, value: float) -> None:
        """Set delay between retries in seconds."""
        self._retry_delay = value

    @property
    def api_key(self: t.Self) -> str | None:
        """Get API key for authentication."""
        return self._api_key

    @api_key.setter
    def api_key(self: t.Self, value: str | None) -> None:
        """Set API key for authentication."""
        self._api_key = value



"""Interface to pandas pddatareader.

:seealso: https://pandas-datareader.readthedocs.io/en/latest/

"""

__author__ = "Jerome Lecomte"
__license__ = "MIT"


import datetime as dt
import typing as t

import pandas_datareader as pdr

import finquotes as fq

SOURCE = "finquotes.datareader.stooq"


class PandasDataReaderFeed(fq.NetworkConfigMixin, fq.AbstractHistoricalFeed, fq.AbstractCloseFeed):
    """Historical close prices from pandas datareader."""

    def __init__(
        self: t.Self,
        as_of_date: dt.date | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Initialize the feed."""
        super().__init__(name=__name__, source=SOURCE, **kwargs)
        self.as_of_date = as_of_date or fq.today()

    def subsource(self: t.Self) -> str:
        """Remove the finquotes.datareader. part of the source."""
        return SOURCE[len("finquotes.datareader.") :]

    def fetch_hist_prices(
        self: t.Self,
        symbol: str,
        end_date: dt.date | None = None,
        begin_date: dt.date | None = None,
        frequency: str | None = None,
        splits: t.Sequence[fq.Split] | None = None,
        distributions: t.Sequence[fq.Distribution] | None = None,
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch historical prices."""
        fq.log.debug(
            f"fetch_hist_prices({symbol}, {begin_date}, {end_date}, {frequency}, {splits}, {distributions})",
        )
        df = pdr.data.DataReader(symbol, self.subsource(), begin_date, end_date)
        for date, row in df.iterrows():
            yield fq.Price(
                symbol=symbol,
                date=date.to_pydatetime().date(),
                close=fq.to_decimal(row["Close"]),
                open_=fq.to_decimal_or_none(row["Open"]),
                high=fq.to_decimal_or_none(row["High"]),
                low=fq.to_decimal_or_none(row["Low"]),
                volume=fq.to_int_or_none(row["Volume"]),
                source=self.source,
            )

    def fetch_close_prices(
        self: t.Self,
        symbols: t.Sequence[str],
        date: dt.date | None = None,
        **kwargs: t.Any,  # noqa: ARG002
    ) -> t.Generator[fq.Price, None, None]:
        """Fetch close prices."""
        df = pdr.data.DataReader(symbols, self.subsource(), date, date).stack(
            future_stack=True,
        )
        for (as_of_date, symbol), row in df.iterrows():
            yield fq.Price(
                symbol=symbol,
                date=as_of_date.to_pydatetime().date(),
                close=fq.to_decimal(row["Close"]),
                open_=fq.to_decimal_or_none(row["Open"]),
                high=fq.to_decimal_or_none(row["High"]),
                low=fq.to_decimal_or_none(row["Low"]),
                volume=fq.to_int_or_none(row["Volume"]),
                source=self.source,
            )


def build_historical_feed(*args: t.Any, **kwargs: t.Any) -> PandasDataReaderFeed:
    """Forward to build_feed."""
    return PandasDataReaderFeed(*args, **kwargs)


def build_close_feed(*args: t.Any, **kwargs: t.Any) -> PandasDataReaderFeed:
    """Forward to build_feed."""
    return PandasDataReaderFeed(*args, **kwargs)
